##############################################################################
#    This file is part of mdciao.
#    
#    Copyright 2020 Charité Universitätsmedizin Berlin and the Authors
#
#    Authors: Guillermo Pérez-Hernandez
#    Contributors:
#
#    mdciao is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    mdciao is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with mdciao.  If not, see <https://www.gnu.org/licenses/>.
##############################################################################

import requests as _requests

from mdtraj import load_pdb as _load_pdb
from tempfile import NamedTemporaryFile as _NamedTemporaryFile
from shutil import copy as _copy, copyfileobj as _copyfileobj

from urllib.error import HTTPError as _HTTPError, URLError as _URLError
from urllib.request import urlopen as _urlopen

def _url2json(url,timeout,verbose):
    r""" Wraps around :obj:`requests.get` and tries to do :obj:`requests.Response.json`

    Parameters
    ----------
    url : str
    timeout : int
        seconds
    verbose : bool

    Returns
    -------
    dict

    """
    if verbose:
        print("Calling %s..." % url, end="")
    a = _requests.get(url, timeout=timeout)
    if verbose:
        print("done!")
    try:
        json = a.json()
    except:
        json = ValueError('Could not create a json out of  %s\n'
                          'Please check (e.g. via browser) if that is a valid url'%url)
    return json

def pdb2ref(pdb, url="https://data.rcsb.org/rest/v1/core/entry/",
            timeout=5):
    r"""
    Print the primary citation of a pdb code via web lookup

    The citation object is returned as a dict,
    check these links for more info
     * https://data.rcsb.org/index.html#data-api
     * https://data.rcsb.org/data-attributes.html

    Parameters
    ----------
    pdb : str
        four-leter pdb-code
    url : str
        the base URL for the look-ups, default is 'https://data.rcsb.org/rest/v1/core/entry/'
    timeout : int
        passed to the :obj:`requests.get` API,
        'How many seconds to wait for the server to send data
        before giving up'

    Returns
    -------
    ref : dict
        Whatever is contained in "rcsb_primary_citation"
        of https://data.rcsb.org/data-attributes.html
    """
    url = url.strip("/")+"/" + pdb.strip("/")
    ref = _url2json(url,timeout,verbose=False)
    try:
        ref = ref["rcsb_primary_citation"]
    except KeyError:
        print(url,":", ref["message"])
        return

    print("Please cite the following 3rd party publication:")
    print(" * "+ref["title"])
    print("   "+ref["rcsb_authors"][0],"et al.,", ref["rcsb_journal_abbrev"], ref["year"])
    print("   https://doi.org/"+ref["pdbx_database_id_doi"])

    return ref

def pdb2traj(code,
             filename=None,
             verbose=True,
             url="https://files.rcsb.org/download/",
             ):
    r""" Return a :obj:`~mdtraj.Trajectory` from a four-letter PDB code via RCSB PDB lookup

    Thinly wraps around :obj:`mdtraj.load_pdb`, printing the corresponding citation.
    Will return None if lookup fails

    Parameters
    ----------
    code : str
        four-letter code, e.g. 3SN6
    filename : str, default is None
        if str, save to this file,
        eventually overwriting
    verbose : bool, default is False
    url : str, default is 'https://files.rcsb.org/download'
        base URL for lookups

    Returns
    -------
    traj : :obj:`~mdtraj.Trajectory` or None

    """
    url1 = "%s/%s.pdb" % (url.strip("/"),code.strip("/"))

    #TODO use print_v elsewhere
    print_v = lambda s, **kwargs: [print(s, **kwargs) if verbose else None][0]
    print_v("Checking %s" % url1, end=" ...", flush=True)
    geom = None
    # From https://docs.python.org/3/howto/urllib2.html#fetching-urls
    try:
        with _urlopen(url1) as response:
            if response.status == 200:
                with _NamedTemporaryFile(delete=True) as tmp_file:
                    _copyfileobj(response, tmp_file)
                    geom = _load_pdb(tmp_file.name)
                    if filename is not None:
                        print_v("Saving to %s..." % filename, end="", flush=True)
                        if filename.lower().endswith(".pdb"):
                            _copy(tmp_file.name, filename)
                        else:
                            geom.save(filename)
                        print_v(filename)
            else:
                raise _URLError(response.reason,filename=url1)


    except (_HTTPError, _URLError) as e:
        print(url1, ":", e)

    if geom is not None:
        pdb2ref(code)

    return geom



r"""
Guess and manipulate fragments, i.e., sub-regions
of molecular topologies. A fragment is just an iterable
of integers representing the residue indices belonging to it.

Methods not documented here are either private or will be soon,
please note the API is not yet stable.

.. currentmodule:: mdciao.fragments

.. autosummary::
   :toctree: generated/
   :nosignatures:

    get_fragments
    fragment_slice
    print_frag
    overview
    match_fragments
    assign_fragments


"""
from .fragments import *
from .fragments import _get_fragments_by_jumps_in_sequence


##############################################################################
#    This file is part of mdciao.
#    
#    Copyright 2020 Charité Universitätsmedizin Berlin and the Authors
#
#    Authors: Guillermo Pérez-Hernandez
#    Contributors:
#
#    mdciao is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    mdciao is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with mdciao.  If not, see <https://www.gnu.org/licenses/>.
##############################################################################

from json import load as _jsonload, JSONDecodeError as _JSONDecodeError
from os.path import splitext as _psplitext, split as _psplit
import numpy as _np
from copy import deepcopy as _dcopy

import mdciao.fragments as _mdcfrg
import mdciao.utils as _mdcu

def x2site(site, fmt="AAresSeq"):
    """
    Return a site dictionary from a dict or an ascii file

    Parameters
    ----------
    site : str or dict
        Path to the ascii file. If the file isn't
        a json file, or doesn't contain the 'name'-field,
        the filename itself will be used as 'name' in the output
        If a dict is passed, it's checked that the dictionary
        has the needed keys to function as a site.
        See :obj:`mdciao.sites` for more info on how sites
        are defined.
    fmt: str, default is "AAresSeq"
        The expected format in case of reading a file.
        Only used when reading a file.

    Returns
    -------
    site : dict
        Keys are:
         * pairs
         * n_pairs
         * name
        site["pairs"] is itself a dictionary
        valued with a list of pairs,
        e.g ["L394","K270"] or [[10,20],[300-4]]],
        depending on the type input

    """
    if isinstance(site, dict):
        idict = _dcopy(site)
    else:
        try:
            with open(site, "r") as f:
                idict = _jsonload(f)
        except _JSONDecodeError as e:
            idict = dat2site(site,fmt=fmt)

    try:
        bondtype = list(idict["pairs"].keys())
        assert len(bondtype)==1 and bondtype[0] in ["AAresSeq","residx"]
        bondtype = bondtype[0]
        pairs = idict["pairs"][bondtype]
        if isinstance(pairs[0][0],str):  # can only be str separated by "-"
            _pairs = []
            for item in pairs:
                if item[0].strip() != '#':
                    if "-" in item and isinstance(item,str):
                        _pairs.append(item.split("-"))
                    elif not isinstance(item, str) and len(item)==2:
                        _pairs.append(item)
                    else:
                        raise ValueError("Can't understand %s"%item)
                idict["pairs"][bondtype] = _pairs

            if bondtype=="residx":
                idict["pairs"][bondtype] = [[int(pp) for pp in pair] for pair in  idict["pairs"][bondtype]]
        else:
            assert all([len(bond)==2 for bond in pairs]),pairs
        idict["n_pairs"] = len(idict["pairs"][bondtype])

    except KeyError:
        print("Malformed file for the site %s:\n%s" % (site,idict))
        raise

    if "name" not in idict.keys():
        if isinstance(site,str):
            idict["name"] = _psplitext(_psplit(site)[-1])[0]
        else:
            raise ValueError("A 'name'-key is mandatory when passing a site dictionary")

    return idict

def sites_to_res_pairs(site_dicts, top,
                       fragments=None,
                       **get_fragments_kwargs,
                       ):
    r"""Return the pairs of res_idxs needed to compute the contacts contained in the input sites.

    The idea is to join all needed pairs of res_idxs
    in one list regardless of what site they come from.

    Note
    ----
    Any residue not found in :obj:`top` is assigned
    a 'None' in the returned :obj:`res_idx_pairs`.

    Parameters
    ----------
    site_dicts : list of dicts
        Anything that :obj:`mdciao.sites.x2site` understands
    top : :obj:`~mdtraj.Topology`
    fragments : list, default is None
        You can pass along fragment definitions so that
        it's easier to de-duplicate any AA in your input.
        Otherwise, these will be created on-the-fly
        by :obj:`mdciao.fragments.get_fragments`
    get_fragments_kwargs :
        see :obj:`fragments.get_fragments`

    Returns
    -------
    res_idxs_pairs : 2D np.ndarray
        Unique residue pairs contained in the :obj:`site_dicts`,
        expressed as residue indices of :obj:`top`
        [0,1] is considered != [0,1]. Any residues that
        couldn't be found will appear as 'None'
    site_maps : list
        For each site, a list with the indices of :obj:`res_idxs_pairs`
        that matches the site's pairs in :obj:`res_idxs_pairs`
    """
    res_idxs_pairs = []
    pair2idx = {}
    site_maps = []
    for ii, site in enumerate(site_dicts):
        imap=[]
        for bond_type, bonds in x2site(site)["pairs"].items():
            if bond_type=="AAresSeq":
                if fragments is None:
                    fragments = _mdcfrg.get_fragments(top, **get_fragments_kwargs)
                get_pair_lambda =  lambda bond: _mdcu.residue_and_atom.residues_from_descriptors(bond, fragments, top)[0]
            elif bond_type=="residx":
                get_pair_lambda = lambda bond: bond
            for bond in bonds:
                pair = tuple(list(get_pair_lambda(bond))+list(bond))
                if pair not in res_idxs_pairs:
                    res_idxs_pairs.append(pair)
                    pair2idx[pair]=len(res_idxs_pairs)-1
                imap.append(pair2idx[pair])
        site_maps.append(imap)
    #print(site_maps)
    return _np.vstack([pair[:2] for pair in res_idxs_pairs]), site_maps

def discard_empty_sites(ctc_idxs, site_maps, site_list, allow_partial_sites=True):
    r"""
    Helper method to reconfigure the output of :obj:`sites_to_res_pairs`
    discarding 'None'-entries

    Parameters
    ----------
    ctc_idxs : list
        List of residue index pairs
    site_maps : list
        List of len :obj:`sites`. Each item
        is itself a list, s.t. :obj:`site_maps[ii]`
        contains the list of indices of pairs
        in :obj:`ctc_idxs` that belong to the ii-th site
    site_list : list
        List containing site definitions
    allow_partial_sites : bool, default is True
        If False, a single 'None' is enough to
        eliminate the entire site. Default
        is to prune the site definitions
        of the 'None' values but keep
        the ones that work

    Returns
    -------
    out_ctc_idxs : list
        Like :obj:`ctc_idxs` but with the
        non-None entries only
    out_site_maps : list
        Like :obj:`site_maps` but with
        updated indices, both of
        the residue pairs in :obj:`site_maps`
        and of the sites in :obj:`out_sites`
        non-None entries
    out_sites : list
        The sites that weren't eliminated.
    discarded : dict
        Keys are "partial" and "full" indicating
        the indices of the sites in :obj:`site_list`
        where either partial deletion took place
        or the full site was discarded completely
    """
    discarded = {}
    if allow_partial_sites:
        gets_eliminated = lambda site : all([None in pair for pair in site["pairs"]["residx"]])
    else:
        gets_eliminated = lambda site : any([None in pair for pair in site["pairs"]["residx"]])

    out_sites = [{"name": isite["name"],
                  "pairs": {"residx": [list(ctc_idxs[ii]) for ii in site_maps[ii]]}}
                 for ii, isite in enumerate(site_list)]

    out_sites = {key: isite for key, isite in enumerate(out_sites) if not gets_eliminated(isite)}
    [isite["pairs"].__setitem__("residx", [pair for pair in isite["pairs"]["residx"] if None not in pair]) for isite in out_sites.values()]

    discarded["partial"] = [ii for ii, isite in out_sites.items() if len(isite["pairs"]["residx"])<len(site_maps[ii])]
    discarded["full"] = [ii for ii in range(len(site_list)) if ii not in out_sites.keys()]

    out_sites = {key: isite for key, isite in out_sites.items() if len(isite["pairs"]["residx"])>0}
    [isite.__setitem__("n_pairs", len(isite["pairs"]["residx"])) for isite in out_sites.values()]
    out_ctc_idxs, out_site_maps = sites_to_res_pairs(out_sites.values(), None)  # we know we won't need top or frags, can just parse None
    return out_ctc_idxs, out_site_maps, list(out_sites.values()), discarded

def site2str(site):
    r""" Produce a printable str for sitefile (json) or site-dict"""
    if isinstance(site,str):
        return site
    elif isinstance(site,dict):
        assert "name" in site.keys()
        return "site dict with name '%s'"%site["name"]
    else:
        raise ValueError("What is this site %s? Only dicts or files are accepted"%site)

def dat2site(dat,comment="#",
             fmt="AAresSeq"):
    r""" Read a non-json (.dat, .txt...) file and turn it into a site dictionary

    The expected format is something like
    # interesting-contacts
    R38-H387
    GDP-R199
    GDP-R201

    The title line is optional, if present,
    will be used as name

    Parameters
    ----------
    dat : str,
        path to a file
    comment : str, default is "#"
        Ignore lines starting with
        the characters in this string
    fmt: str, default is "AAresSeq"
        The expected format of the file.
        'AAresSeq' means that the pairs
        are understood as AA-names
        followed by a sequence index:
        "GLU30-ARG131".

    -------
    site : a dictionary
     Same format as return of :obj:`x2site`
     "name" will be whatever :obj:`dat` was,
     without the extension
    """
    with open(dat,"r") as f:
        lines = f.read().splitlines()
    offset = 0
    name = _psplitext(_psplit(dat)[-1])[0]
    if lines[0].strip(" ").startswith("#"):
        name = lines[0].split("#")[1].strip(" ")
        offset +=1
    assert fmt in ["AAresSeq", "residx"], NotImplementedError("Only [AAresSeq, residx] are implemented for 'fmt' at the moment, can't do '%s'"%(fmt))
    site={"pairs":{fmt:[]}}
    for ii, line in enumerate(lines[offset:]):
        if line.strip(" ")[0] not in comment:
            assert line.count("-")==1, ValueError("The contact descriptor has to contain one (and just one) '-', got %s instead (%u-th line)"%(line, ii+1))
            site["pairs"][fmt].append(line)
            if fmt=="AAresSeq" and not any([chr.isalpha() for chr in line.replace("-","")]):
                raise ValueError("This can't be a %s line %s, are you sure you don't mean fmt=%s"%(fmt,line,"residx"))
            elif fmt=="residx" and not all([chr.isdigit() for chr in line.replace("-","")]):
                raise ValueError("This can't be a %s line %s are you sure you don't mean fmt=%s"%(fmt,line,"AAresSeq"))
    site["name"]=name
    return site




##############################################################################
#    This file is part of mdciao.
#    
#    Copyright 2020 Charité Universitätsmedizin Berlin and the Authors
#
#    Authors: Guillermo Pérez-Hernandez
#    Contributors:
#
#    mdciao is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    mdciao is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with mdciao.  If not, see <https://www.gnu.org/licenses/>.
##############################################################################

import numpy as _np
import mdtraj as _md
from mdtraj.core.residue_names import _PROTEIN_RESIDUES
import mdciao.utils as _mdcu
from pandas import unique as _pandas_unique

_allowed_fragment_methods = ['chains',
                             'resSeq',
                             'resSeq+',
                             'lig_resSeq+',
                             'bonds',
                             'resSeq_bonds',
                             "None",
                             ]


def print_fragments(fragments, top, max_lines=40, **print_frag_kwargs):
    """Inform about fragments, very thinly wrapping around :obj:`print_frag`

    Parameters
    ----------
    fragments : dict or list
        Iterable with the sets of residue indexes
    top : :obj:`~mdtraj.Topology`
    max_lines : int, default is 40
        Maximum number of lines to print. E.g. if 40,
        first 20 and last 20 will be printed
    print_frag_kwargs : opt, keyword args for :obj:`print_frag`

    Returns
    -------
    frag_list : list
        List of the strings that
        are being printed, for further use

    """
    if isinstance(fragments,list):
        _fragments = {ii:val for ii, val in enumerate(fragments)}
    else:
        _fragments = {key:val for key, val in fragments.items()}
    frag_list = []
    for ii, iseg in _fragments.items():
        frag_list.append(print_frag(ii, top, iseg, return_string=True, **print_frag_kwargs))
    n = _np.round(max_lines/2).astype(int)
    if n * 2 < len(frag_list):
        frag_list = frag_list[:n] + ["...[long list: omitted %u items]..." % (len(frag_list) - 2 * n)] + frag_list[-n:]
    print("\n".join(frag_list))
    return frag_list

def print_frag(frag_idx, top, fragment, fragment_desc='fragment',
               idx2label=None,
               return_string=False,
               resSeq_jumps=True,
               label_width=10,
               **print_kwargs):
    """Pretty-printing of fragments of an :obj:`mtraj.topology`

    Parameters
    ----------
    frag_idx: int or str
        Index or name of the fragment to be printed
    top: :obj:`~mdtraj.Topology` or string
        Topology or string (=AA sequence) to "grab"
        the residue-names from when informing
    fragment: iterable of indices
        The fragment in question, with zero-indexed residue indices
    fragment_desc: str, default is "fragment"
        How to call the fragments, e.g. segment, block, monomer, chain
    idx2label : iterable or dictionary
        Pass along any consensus labels here
    resSeq_jumps : bool, default is True
        Inform whether the fragment contains jumps in the resSeq
    label_width : int, default is 10
        The width in characters given
        to the label descriptor. You can set this to zero
        if you know :obj:`idx2label` is None for all
        printed lines
    return_string: bool, default is False
        Instead of printing, return the string
    print_kwargs:
        Optional keyword arguments to pass to the print function, e.g. "end=","" and such

    Returns
    -------
    None or str, see return_string option

    """
    maplabel_first, maplabel_last = "", ""
    labfmt = "%-"+"%us"%label_width
    try:
        if idx2label is not None:
            maplabel_first = _mdcu.str_and_dict.choose_options_descencing([idx2label[fragment[0]]],
                                                                                       fmt="@%s")
            maplabel_last = _mdcu.str_and_dict.choose_options_descencing([idx2label[fragment[-1]]],
                                                                                      fmt="@%s")
        if isinstance(top,_md.Topology):
            rfirst, rlast = [top.residue(ii) for ii in [fragment[0], fragment[-1]]]

        elif isinstance(top,str):
            rfirst, rlast = [top[ii] for ii in [fragment[0], fragment[-1]]]
        rfirst_index, rlast_index = [fragment[0],fragment[-1]]

        labfirst = "%8s%-10s" % (rfirst, maplabel_first)
        lablast = "%8s%-10s" % (rlast, maplabel_last)
        istr = "%s %6s with %6u AAs %8s%s (%6u) - %8s%s (%-6u) (%s) " % \
               (fragment_desc, str(frag_idx), len(fragment),
                #labfirst,
                rfirst, labfmt%maplabel_first,
                rfirst_index,
                #lablast,
                rlast, labfmt%maplabel_last,
                rlast_index,
                str(frag_idx))

        if isinstance(top,_md.Topology) and  rlast.resSeq - rfirst.resSeq != len(fragment) - 1:
            # print(ii, rj.resSeq-ri.resSeq, len(iseg)-1)
            istr += ' resSeq jumps'
    except:
        print(fragment)
        raise
    if return_string:
        return istr
    else:
        print(istr, **print_kwargs)

def get_fragments(top,
                  method='lig_resSeq+',
                  fragment_breaker_fullresname=None,
                  atoms=False,
                  verbose=True,
                  join_fragments=None,
                  maxjump = 500,
                  salt=["Na+","Cl-","Na","Cl"],
                  water=True,
                  **kwargs_residues_from_descriptors):
    """
    Group residues of a molecular topology into fragments using different methods.

    Water and ions get their own fragment by default except for the methods
    None, chains, and any method involving bonds

    Parameters
    ----------
    top : :obj:`~mdtraj.Topology` or str
        When str, path to filename

    method : str, default is 'lig_resSeq+'
        The method passed will be the basis for creating fragments. Check the following options
        with the example sequence

            "…-A27,Lig28,K29-…-W40,D45-…-W50,CYSP51,GDP52"

        - 'resSeq'
            breaks at jumps in resSeq entry:

            […A27,Lig28,K29,…,W40],[D45,…,W50,CYSP51,GDP52]

        - 'resSeq+'
            breaks only at negative jumps in resSeq:

            […A27,Lig28,K29,…,W40,D45,…,W50,CYSP51,GDP52]

        - 'bonds'
            breaks when residues are not connected by bonds, ignores resSeq:

            […A27][Lig28],[K29,…,W40],[D45,…,W50],[CYSP51],[GDP52]

            notice that because phosphorylated CYSP51 didn't get a
            bond in the topology, it's considered a ligand

        - 'resSeq_bonds'
            breaks at resSeq jumps and at missing bonds
        - 'lig_resSeq+'
            Like resSeq+ but put's any non-AA residue into it's own fragment.
            […A27][Lig28],[K29,…,W40],[D45,…,W50,CYSP51],[GDP52]
            Also check :obj:`maxjump`
        - 'chains'
            breaks into chains of the PDB file/entry
        - None or 'None'
            all residues are in one fragment, fragment 0
    fragment_breaker_fullresname : list
        list of full residue names. Example [GLU30] will be used to break fragments,
        so that [R1, R2, ... GLU30,...R10, R11] will be broken into [R1, R2, ...], [GLU30,...,R10,R11]
    atoms : boolean, optional
        Instead of returning residue indices, return atom indices
    join_fragments : list of lists
        After getting the fragments with :obj:`method`,
        join these fragments again. The use case are hard
        cases where no method gets it right and some post-processing
        is needed.
        Duplicate entries in any inner list will be removed.
        One fragment idx cannot appear in more than one inner list,
        otherwise an exception is thrown
    verbose : boolean, optional
        Be verbose
    salt : list, default is ["Na+","Cl+", "NA","CL"]
        Residues that match these residue names and
        have only one atom will be put together
        in the last fragment. Use salt = []
        to deactivate. Doesn't apply for methods
        involving bonds or None and chains
    water : bool, default is True
        Put water on its own fragment.
        Doesn't apply for methods
        involving bonds or None and chains
    maxjump : int or None, default is 500
        The maximum allowed positive sequence-jump
        in the 'resSeq+' methods, i.e. don't
        join ALA500 with GLU551 even though
        the jump in sequence is positive
        None means no limit for positive jumps
    kwargs_residues_from_descriptors : optional
        additional arguments, see :obj:`~mdciao.residue_and_atom.residues_from_descriptors`

    Returns
    -------
    List of integer arrays
        Each array within the list has the residue indices of each fragment.
        These fragments do not have overlap. Their union contains all indices

    """

    _assert_method_allowed(method)
    salt = [ss.lower() for ss in salt]
    if isinstance(top, str):
        top = _md.load(top).top

    # Auto detect fragments by resSeq
    fragments_resSeq = _get_fragments_by_jumps_in_sequence([rr.resSeq for rr in top.residues])[0]

    if method=="resSeq":
        fragments = fragments_resSeq
    elif method=='resSeq_bonds':
        residue_bond_matrix = _mdcu.bonds.top2residue_bond_matrix(top, verbose=False,
                                                      force_resSeq_breaks=True)
        fragments = _mdcu.bonds.connected_sets(residue_bond_matrix)
    elif method=='bonds':
        residue_bond_matrix = _mdcu.bonds.top2residue_bond_matrix(top, verbose=False,
                                                      force_resSeq_breaks=False)
        fragments = _mdcu.bonds.connected_sets(residue_bond_matrix)
        fragments = [fragments[ii] for ii in _np.argsort([fr[0] for fr in fragments])]
    elif method == "chains":
        fragments = [[rr.index for rr in ichain.residues] for ichain in top.chains]
    elif method == "resSeq+":
        fragments = _get_fragments_resSeq_plus(top, fragments_resSeq,maxjump=maxjump)
    elif method == "lig_resSeq+":
        fragments = _get_fragments_resSeq_plus(top, fragments_resSeq,maxjump=maxjump)
        lig_cands = _np.unique([top.atom(aa).residue.index for aa in top.select("not protein and not water")])
        for ii in lig_cands:
            rr = top.residue(ii)
            if rr.name[:3] in _PROTEIN_RESIDUES or rr.name.lower() in salt:
                continue
            frag_idx = _mdcu.lists.in_what_fragment(rr.index,fragments)
            if len(fragments[frag_idx])>1:
                list_for_removing=list(fragments[frag_idx])
                list_for_removing.remove(rr.index)
                fragments[frag_idx]=_np.array(list_for_removing)
                fragments.append([rr.index])
    # TODO check why this is not equivalent to "bonds" in the test_file
    elif method == 'molecules':
        raise NotImplementedError("method 'molecules' is not fully implemented yet")
        #fragments = [_np.unique([aa.residue.index for aa in iset]) for iset in top.find_molecules()]
    elif method == 'molecules_resSeq+':
        raise NotImplementedError("method 'molecules_resSeq+' is not fully implemented yet")
        """
        _molecules = top.find_molecules()
        _fragments = [_np.unique([aa.residue.index for aa in iset]) for iset in _molecules]
        _resSeqs =   [[top.residue(idx).resSeq for idx in ifrag] for ifrag in _fragments]
        _negjumps =  [_np.diff(_iresSeq) for _iresSeq in _resSeqs]
        fragments = []
        for ii, (ifrag, ijumps) in enumerate(zip(_fragments, _negjumps)):
            if all(ijumps>0):
                fragments.append(ifrag)
            else:
                offset = 0
                for kk, jj in enumerate(_np.argwhere(ijumps<0)):
                    print(ii, jj)
                    jj = int(jj)
                    #print(top.residue(ifrag[jj]), top.residue(ifrag[jj+1]))
                    offset+=jj
                    fragments.append(ifrag[:jj+1])
                    fragments.append(ifrag[jj+1:])
                    if kk>1:
                        raise Exception
        #print(_negjumps)
        """
    elif str(method).lower() == "none":
        fragments = [_np.arange(top.n_residues)]

    if "resSeq" in str(method) and "bond" not in str(method):
        if water:
            fragments = _dry_fragments(fragments, top)
        fragments = _bland_fragments(fragments, top, salt)

    fragments = [fragments[ii] for ii in _np.argsort([ifrag[0] for ifrag in fragments])]

    # Inform of the first result
    if verbose:
        print("Auto-detected fragments with method '%s'"%str(method))
        print_fragments(fragments,top,label_width=0)
    # Join if necessary
    if join_fragments is not None:
        fragments = _mdcu.lists.join_lists(fragments, join_fragments)
        print("Joined Fragments")
        for ii, iseg in enumerate(fragments):
            print_frag(ii, top, iseg)

    if fragment_breaker_fullresname is not None:
        if isinstance(fragment_breaker_fullresname,str):
            fragment_breaker_fullresname=[fragment_breaker_fullresname]
        for breaker in fragment_breaker_fullresname:
            #todo use _break_fragments
            residxs, fragidx = _mdcu.residue_and_atom.residues_from_descriptors(breaker, fragments, top,
                                                           **kwargs_residues_from_descriptors)
            idx = residxs[0]
            ifrag = fragidx[0]
            if idx is None:
                print("The fragment breaker %s appears nowhere" % breaker)
                # raise ValueError
            else:
                idx_split = _np.argwhere(idx ==_np.array(fragments[ifrag])).squeeze()
                #print('%s (index %s) found in position %s of frag %s%s' % (breaker, idx, idx_split, ifrag, fragments[ifrag]))
                subfrags = [fragments[ifrag][:idx_split], fragments[ifrag][idx_split:]]
                print("New fragments after breaker %s:" % breaker)
                if idx_split>0:
                    #print("now breaking up into", subfrags)
                    fragments = fragments[:ifrag] + subfrags + fragments[ifrag + 1:]
                    for ii, ifrag in enumerate(fragments):
                        print_frag(ii, top, ifrag)
                else:
                    print("Not using it since it already was a fragment breaker in frag %s"%ifrag)
            print()
            # raise ValueError(idx)

    if not atoms:
        return fragments
    else:
        return [_np.hstack([[aa.index for aa in top.residue(ii).atoms] for ii in frag]) for frag in fragments]

def _break_fragments(breakers, fragments):
    r"""
    Given a list of fragment breakers, break existing fragments further into sub-fragments

    The break is so that j appearing in fragment [a, b, ...,i,j,...z] will
    generate [a, b, ...,i],[j,...z]

    Parameters
    ----------
    breakers : iterable of idxs
        These indices will force
        a break in the :obj:`fragments`,
        generating sub-fragments.
    fragments : iterable of iterables
        The fragment definitions

    Returns
    -------
    fragments : list
    """
    _mdcu.lists.assert_no_intersection(fragments)
    for idx in _np.unique(breakers):
        ifrag = _mdcu.lists.in_what_fragment(idx, fragments)
        if ifrag is not None:
            idx_split = _np.flatnonzero(idx == _np.array(fragments[ifrag]))[0]
            subfrags = [fragments[ifrag][:idx_split], fragments[ifrag][idx_split:]]
            if idx_split > 0:
                # print("now breaking up into", subfrags)
                fragments = fragments[:ifrag] + subfrags + fragments[ifrag + 1:]
    return fragments

#TODO use this throught the code instead of the python explicit way
def fragment_slice(traj : _md.Trajectory, fragments, keys_or_idxs=None):
    r"""

    Slice a geometry using arbitrary fragment definitions, a la :obj:`mdtraj.Trajectory.atom_slice`

    Note
    ----
    Regardless of the order in which the selection
    is done (e.g. keys_or_idxs=[1,0]) the returned
    :obj:`sliced_traj` will always be in ascending
    order atoms as they appear in :obj:`traj`
    Parameters
    ----------
    traj : :obj:`mdtraj.Trajectory`
        The trajectory to slice
    fragments : list or dict
        The fragment definitions as residue indices.
        Can be as a list or a as a dict, e.g. the output of
        :obj:`mdciao.fragments.get_fragments` (list) or
        :obj:`mdciao.nomenclature.LabelerGPCR.top2frags` (dict)
    keys_or_idxs : iterable or None
        The keys or indices of the
        fragments to slice to, i.e.
        to keep. If None, all
        all fragments are used
        as a selection

    Returns
    -------
    sliced_traj : :obj:`mdtraj.Trajectory`
        A copy of :obj:`traj` only with the
        atoms present in the selected fragments

    """

    if keys_or_idxs is None:
        if isinstance(fragments,dict):
            keys_or_idxs = list(fragments.keys())
        else:
            keys_or_idxs = _np.arange(len(fragments))

    _fragments = _np.hstack([fragments[idx] for idx in keys_or_idxs])

    return traj.atom_slice([aa.index for aa in traj.top.atoms if aa.residue.index in _fragments])

def _dry_fragments(fragments, top):
    r"""
    Remove water molecules from :obj:`fragments` and append them at the end as their own fragment(s)

    Water is selected with top.select("water") (check https://www.mdtraj.org/1.9.5/atom_selection.html)

    Parameters
    ----------
    fragments : list of ints
    top : :obj:`~mdtraj.Topology`

    Returns
    -------
    dry_fragments : list
        The original :obj:`fragments` except
        the water molecules have been put into
        their own fragment(s) at the end of the list.
        If the water molecules are not contiguous in
        their serial residue-indices (not resSeq),
        they get split into contiguous fragments
    """
    waters = _np.unique([top.atom(aa).residue.index for aa in top.select("water")])
    if len(waters)>0:
        return _mdcu.lists.remove_from_lists(fragments, waters)+_get_fragments_by_jumps_in_sequence(waters)[1]
    else:
        return fragments

def _bland_fragments(fragments, top, salt):
    r"""
    Remove salt from :obj:`fragments` and append them at the end as their own fragment(s)

    Parameters
    ----------
    fragments : list of ints
    top : :obj:`~mdtraj.Topology`
    salt : list
        Ions will be matched using
        residue.name in :obj:`salt`
        and ensuring residue.n_atoms == 1

    Returns
    -------
    bland_fragments : list
        The original :obj:`fragments` except
        the ions have been put into
        their own fragment(s) at the end of the list.
        If the ions are not contiguous in
        their serial residue-indices (not resSeq),
        they get split into contiguous fragments
    """
    salt = [ss.lower() for ss in salt]
    ion_cands = _np.unique([top.atom(aa).residue.index for aa in top.select("not protein and not water")])
    ion_cands = [ii for ii in ion_cands if top.residue(ii).name.lower() in salt and top.residue(ii).n_atoms==1]
    if len(ion_cands)>0:
        return _mdcu.lists.remove_from_lists(fragments, ion_cands) + _get_fragments_by_jumps_in_sequence(ion_cands)[1]
    else:
        return fragments


def match_fragments(seq0, seq1,
                    frags0=None,
                    frags1=None,
                    probe=None,
                    verbose=False,
                    shortest=3):
    r"""
    Align fragments of seq0 and seq1 pairwise and return a matrix of scores.

    The score is the absolute number matches between
    two fragments. Depending on how informed
    the user is about the topologies, their fragments,
    their similarities, and what the user is trying
    to do, this absolute measure can be just right or
    be highly misleading, e.g:
     * two fragments of ~500 AAs each can score
       20 matches "easily", without this being meaningful
     * two fragments of 11 AAs each having 10 matches
       between them are almost identical
     * however, in absolute terms, the first case has a
       higher score

    If you know what you're doing, you can specify
    which one of the sequences is the :obj:`probe`, s.t.
    the score is divided by the length the fragment of
    the probe. E.g., if :obj:`probe` =1, it means
    that you are interested in finding out if
    fragments of :obj:`seq1` appear in fragments of :obj:`seq0`,
    (the 'target' sequence), regardless of how long
    the target fragments are. The score is
    then normalized to 1, where 1 means you
    found the entire probe fragment in the target fragment,
    no matter how long the probe or the target were.


    Parameters
    ----------
    seq0 : str or :obj:`~mdtraj.Topology`
    seq1 : str or :obj:`~mdtraj.Topology`
    frags0 : list or None, default is None
        If None, :obj:`get_fragments` will
        be called with the default options
        to generate a fragment list.
    frags1 : list or None, default is None
        If None, :obj:`get_fragments` will
        be called with the default options
        to generate a fragment list.
    probe : int, default is None
        If None, scores are absolute numbers.
        If 0, the scores are divided by
        the seq0's fragment length. If 1,
        by seq1's fragment length. In
        these cases, the score is
        always between 0 and 1,
        regardless how long the probe
        and the target fragments are.
    shortest : int, default is 3
        Fragments of len < :obj:`shortest`
        won't produce a score but a np.NaN,
        s.t. the :obj:`score` doesn't get
        highjacked by very small :obj:`probe`
        fragments, which will always yield
        relative good scores.
        Absolute scores (:obj:`probe` = None)
        are not affected by this.
    verbose : bool, default is False
        Be verbose, affects all methods
        called by the this method as well.

    Returns
    -------
    score : 2D np.ndarray of shape(len(frags0),len(frags1))
        Will be between 0 and 1 if a :obj:`probe`
        is specified
    frags0 : list
        The fragments that were either provided
        or generated on the fly. Their indices
        are the row-indices of :obj:`score`
    frags1 : list
        The fragments that were either provided
        or generated on the fly. Their indices
        are the row-indices of :obj:`score`
    """

    frags = []
    for iseq,ifrg in zip([seq0, seq1],
                         [frags0, frags1]):
        if ifrg is None:
            if isinstance(iseq,_md.Topology):
                ifrg = get_fragments(iseq,verbose=verbose)
            else:
                assert isinstance(iseq,str)
                ifrg = [_np.arange(len(iseq))]
        frags.append(ifrg)

    score = _np.zeros((len(frags[0]),len(frags[1])),dtype=float)
    for ii, ifrag in enumerate(frags[0]):
        for jj,jfrag in enumerate(frags[1]):
            df = _mdcu.sequence.align_tops_or_seqs(seq0, seq1,
                                                   seq_0_res_idxs=frags[0][ii],
                                                   seq_1_res_idxs=frags[1][jj],verbose=verbose)[0]
            score[ii,jj] = df["match"].sum()
            if probe is not None:
                if len([ifrag,jfrag][probe])<=shortest:
                    score[ii,jj] = _np.nan
                else:
                    score[ii,jj]/=len([ifrag,jfrag][probe])

            if verbose:
                for tt, itop in enumerate([seq0,seq1]):
                    print_frag(ii, seq0, ifrag, "seq%u fragment"%tt)
                print(score[ii,jj].round(2))
    return score, frags[0], frags[1]

def _get_fragments_by_jumps_in_sequence(sequence,jump=1):
    r"""
    Return the array in :obj:`sequence` chopped into sub-arrays (=fragments)
    where there's a jump in the sequence

    Parameters
    ----------
    sequence : iterable of ints
    jump : int, default is 1
        What is considered a jump in the sequence

    Returns
    -------
    idxs : list
        The idxs of :obj:`sequence` in each fragment
    fragments : list
        :obj:`sequence` chopped into the fragments
    """
    old = sequence[0]
    frag_idxs = [[]]
    frag_elements = [[]]
    for ii, rr in enumerate(sequence):
        delta = _np.abs(rr - old)
        # print(delta, ii, rr, end=" ")
        if delta <= jump:
            # print(delta, ii, rr, "appending")
            frag_idxs[-1].append(ii)
            frag_elements[-1].append(rr)
        else:
            # print(delta, ii, rr, "new")
            frag_idxs.append([ii])
            frag_elements.append([rr])
        # print(fragments[-1])
        old = rr
    return frag_idxs, frag_elements

#TODO combine with the above method, they are pretty redundant
def _get_fragments_resSeq_plus(top, fragments_resSeq,maxjump=None):
    r"""
    Get fragments using the 'resSeq+' method
    Parameters
    ----------
    top
    fragments_resSeq
    maxjump

    Returns
    -------

    """
    to_join = [[0]]
    jump_is_short    = lambda idx1,idx2 :  [True if maxjump is None else idx2-idx1 <= maxjump][0]
    for ii, ifrag in enumerate(fragments_resSeq[:-1]):
        r1 = top.residue(ifrag[-1])
        r2 = top.residue(fragments_resSeq[ii + 1][0])
        if r1.resSeq < r2.resSeq and jump_is_short(r1.resSeq, r2.resSeq):
            to_join[-1].append(ii + 1)
        else:
            to_join.append([ii + 1])

    return _mdcu.lists.join_lists(fragments_resSeq, [tj for tj in to_join if len(tj) > 1])

def overview(topology,
             methods=['all'],
             AAs=None,
             ):

    """
    Prints the fragments obtained by :obj:`get_fragments` for the available methods.

    Optionally, you can pass along a list of residue
    descriptors to be printed after the fragments have
    been shown.

    Parameters
    ----------
    topology :  :obj:`mdtraj.Topology`
    methods : str or list of strings
        method(s) to be used for obtaining fragments
    AAs : list, default is None
        Anything that :obj:`find_AA` can understand

    Returns
    -------
    fragments_out : dict
        The result of the fragmentation schemes keyed
        by their method name

    """

    if isinstance(methods,str):
        methods = [methods]

    if methods[0].lower() == 'all':
        try_methods = _allowed_fragment_methods
    else:
        for method in methods:
            _assert_method_allowed(method)
        try_methods = methods

    fragments_out = {}
    for method in try_methods:
        try:
            fragments_out[method] = get_fragments(topology,
                                                  method=method)
        except Exception as e:
            print("The method %s did not work:"%method)
            print(e)
            print()
        print()

    _mdcu.residue_and_atom.parse_and_list_AAs_input(AAs, topology)

    return fragments_out

def _assert_method_allowed(method):
    assert str(method) in _allowed_fragment_methods, ('input method %s is not known. ' \
                                                      'Know methods are\n%s ' %
                                                      (method, "\n".join(_allowed_fragment_methods)))

def check_if_subfragment(sub_frag, fragname, fragments, top,
                         map_conlab=None,
                         keep_all=False,
                         prompt=True):
    r"""
    Input an iterable of integers representing a fragment and check if
    it clashes with other fragment definitions.

    Return False/True or prompt for a choice in case it is necessary

    Example
    -------
    Let's assume the GPCR-nomenclature tells us that TM6 is [0,1,2,3]
    and we have already divided the topology into fragments
    using :obj:`get_fragments`, with method "resSeq+", meaning
    we have fragments for the receptor,Ga,Gb,Gg

    The purpose is to check whether the GPCR-fragmentation is
    contained in the previous fragmentation:
    * [0,1,2,3] and :obj:`fragments`=[[0,1,2,3,4,6], [7,8,9]]
    is not a clash, bc TM6 is contained in fragments[0]
    * [0,1,2,3] and :obj:`fragments`=[[0,1],[2,3],[4,5,6,7,8]]
    is a clash. In this case the user will be prompted to choose
    which subset of "TM6" to keep:
     * "0": [0,1]
     * "1": [2,3]
     * "0-1" [0,1,2,3]


    Parameters
    ----------
    sub_frag : iterable of integers
        The fragment to be checked if
        subfragment of `fragments`
    fragname : str
    fragments : iterable of iterables of integers
    top : :obj:`~mdtraj.Trajectory` object
    map_conlab : list or dict, default is None
        maps residue idxs to consensus labels
    prompt : bool, default is True
        When False, no prompt is issued,
        and the returned value is a boolean
        whether or not `sub_frag` is actually a sub-fragment (=subset)
        of any one fragment of `fragments`
    Returns
    -------
    answer : 1D numpy array or boolean
        If `prompt` is True and no clashes were found,
        this will contain the same residues as
        :obj:`sub_frag` without prompting the user.
        Otherwise, the user has to input whether
        to leave the definition intact or pick a sub-set
        If `prompt` is False, this is a boolean
        saying whether `sub_frag` is a sub-fragment (subset)
        of any one fragment of `fragments`
    """
    # Get the fragment idxs of all residues in this fragment
    ifrags = [_mdcu.lists.in_what_fragment(idx, fragments) for idx in sub_frag]

    frag_cands = [ifrag for ifrag in _pandas_unique(ifrags) if ifrag is not None]
    if not prompt:
        return len(frag_cands) <= 1

    if len(frag_cands) > 1 and not keep_all:
        # This only happens if more than one fragment is present
        print_frag(fragname, top, sub_frag, fragment_desc='',
                   idx2label=map_conlab)
        print(" Subfragment %s clashes with other fragment definitions,\n"
              " because the residues of %s span over more than 1 fragment:"%(fragname, fragname))
        for jj in frag_cands:
            istr = print_frag(jj, top, fragments[jj],
                              fragment_desc="   input fragment",
                              return_string=True,
                              label_width=0)
            n_in_fragment = len(_np.intersect1d(sub_frag, fragments[jj]))
            if n_in_fragment < len(fragments[jj]):
                istr += "%u residues outside %s" % (len(fragments[jj]) - n_in_fragment, fragname)
            print(istr)
        answr = input("Input what fragment idxs to include into %s  (fmt = 1 or 1-4, or 1,3):" % fragname)
        answr = _mdcu.lists.rangeexpand(answr)
        assert all([idx in ifrags for idx in answr])
        tokeep = _np.hstack([idx for ii, idx in enumerate(sub_frag) if ifrags[ii] in answr]).tolist()
        if len(tokeep) >= len(ifrags):
            raise ValueError("Cannot keep these fragments %s!" % (str(answr)))
        return tokeep
    else:
        return sub_frag

def _fragments_strings_to_fragments(fragment_input, top, verbose=False):
    r"""

    #TODO rename the strings to something else
    Try to understand how the user wants to fragment the topology
    Pretty flexible

    Check also :obj:`rangeexpand` to understand the expressions


    Parameters
    ----------
    fragment_input : list of strings
        Many cases are possible
        * ["consensus"] : user wants to use
        consensus labels such as "TM6" etc.
        Using "resSeq+" heuristic to fragment, and
        returns :obj:`user_wants_consensus` as True,
        this will trigger further prompts
        * [method] : fragment using "method"
        (see :obj:`get_fragments`) and return
        user_wants_consensus as False
        * [['exp1']] : this str represents the
        residues in one fragment (eg. "0-3,5" : 0,1,2,3,5).
        Assume that any missing residues belong the other fragment.
        Return the two fragments and user_wants_consensus as False
        * [["exp1"],
           ["exp2"],
           [...]]
        These strs are the fragments expressed as residue
        indices. Evaluate them and return them. Return
        user_wants_consensus as False
        * exp can also b ["LEU30-GLU40"]
        * None or "None"

    top : :obj:`~mdtraj.Topology`

    Returns
    -------
    fragments_as_residue_idxs, user_wants_consensus

    """
    user_wants_consensus = False
    assert isinstance(fragment_input,list)
    #TODO the following line is untested, the usecase not mentioned in the docs...?
    if len(fragment_input)==1 and isinstance(fragment_input[0],str) and " " in fragment_input[0]:
        fragment_input = fragment_input[0].split(" ")
    #if len(fragment_input)==1 and isinstance(fragment_input[0],str) and "," in fragment_input[0]:
    #    fragment_input=fragment_input[0].split(",")
    if str(fragment_input[0]).lower()=="consensus":
        user_wants_consensus = True
        method = 'resSeq+'
        fragments_as_residue_idxs = get_fragments(top, method='resSeq+',
                                                  verbose=False)
    elif str(fragment_input[0]) in _allowed_fragment_methods:
        method = fragment_input[0]
        fragments_as_residue_idxs = get_fragments(top, method=method,
                                                  verbose=False)
    else:
        method = "user input by residue array or range"
        fragments_as_residue_idxs = []
        temp_fragments=get_fragments(top,verbose=False)
        for fri in fragment_input:
            if not isinstance(fri,str):
                fragments_as_residue_idxs.append(fri)
            else:
                fragments_as_residue_idxs.append(
                    _mdcu.residue_and_atom.rangeexpand_residues2residxs(fri,
                                                                        temp_fragments,
                                                                        top,
                                                                        interpret_as_res_idxs=fri.replace("-","").replace(",","").isnumeric(),
                                                                        extra_string_info="\nThis fragmentation is only for disambiguation purposes:"
                                                                        ))
        if len(fragment_input)==1:
            assert isinstance(fragment_input[0],str)
            method += " with only one fragment provided (all other residues are fragment 2)"
            fragments_as_residue_idxs.append(_np.delete(_np.arange(top.n_residues), fragments_as_residue_idxs[0]))

    _mdcu.lists.assert_no_intersection(fragments_as_residue_idxs, word="fragments")
    for ii, ifrag in enumerate(fragments_as_residue_idxs):
        if not all([aa in range(top.n_residues) for aa in ifrag]):
            print("Fragment %u's definition had idxs outside of "
                  "the geometry (total n_residues %u) that have been deleted %s " % (
                  ii, top.n_residues, set(ifrag).difference(range(top.n_residues))))
            fragments_as_residue_idxs[ii]=[jj for jj in ifrag if jj<top.n_residues]

    if verbose:
        print("Using method '%s' these fragments were found" % method)
        for ii, ifrag in enumerate(fragments_as_residue_idxs):
            print_frag(ii, top, ifrag, label_width=0)

    return fragments_as_residue_idxs, user_wants_consensus

def frag_list_2_frag_groups(frag_list,
                            frag_idxs_group_1=None,
                            frag_idxs_group_2=None,
                            verbose=False):
    r"""
    Automagically find out the user wants to define
    two fragments out of list of fragments. This is used
    by CLTs interface

    Prompt the user when disambiguation is needed.

    Parameters
    ----------
    frag_list : list
        list of fragments, defined as residue indices
    frag_idxs_group_1 : iterable of ints, or str default is None
        When str, it has to be a rangeexpand exprs, e.g. 2-5,10
    frag_idxs_group_2 : iterable of ints, or str default is None
        When str, it has to be a rangeexpand exprs, e.g. 2-5,10

    Returns
    -------
    groups_as_residxs, groups_as_fragidxs

    """

    if verbose:
        for ii, ifrag in enumerate(frag_list):
            print("frag %u: %u-%u"%(ii, ifrag[0],ifrag[-1]))

    if len(frag_list) == 2 and \
            frag_idxs_group_1 is None and \
            frag_idxs_group_2 is None:
        print("Only two fragments detected with no values for frag_idxs_group_1 and frag_idxs_group_2.\n"
              "Setting frag_idxs_group_1=0 and frag_idxs_group_2=1")
        groups_as_fragidxs = [[0], [1]]
        #TODO I don't think i need the check for None in the frag_idxs_groups, right?
    else:
        groups_as_fragidxs = [frag_idxs_group_1, frag_idxs_group_2]
        for ii, ifrag_idxs in enumerate(groups_as_fragidxs):
            if ifrag_idxs is None:
                groups_as_fragidxs[ii] = _mdcu.lists.rangeexpand(input('Input group of fragments '
                                                                     '(e.g. 0,3 or 2-4,6) for group %u: ' % (ii + 1)))
            elif isinstance(ifrag_idxs, str):
                groups_as_fragidxs[ii] = _mdcu.lists.rangeexpand(ifrag_idxs.strip(","))
    groups_as_residxs = [sorted(_np.hstack([frag_list[ii] for ii in iint])) for iint in
                            groups_as_fragidxs]

    return groups_as_residxs, groups_as_fragidxs

def frag_dict_2_frag_groups(frag_defs_dict, ng=2,
                            verbose=False,
                            answers=None):
    r"""
    Input a dictionary of fragment definitions, keyed by
    whatever and valued with residue idxs and prompt
    the user how to re-group them

    It wraps around :obj:`_match_dict_by_patterns`
    under the hood

    TODO: refactor into str_and_dict_utils
    TODO: It will be mostly used with fragments so it's better here for the API? IDK

    Parameters
    ----------
    frag_defs_dict : dict
        Fragment definitions in residue idxs
    ng : int, default is 2
        wanted number of groups
    answers : list, default is None
        List of strings. If provided,
        the items of this list will
        be passed as answers to the prompt
        asking for fragment choice. None and

    Returns
    -------
    groups_as_residue_idxs, groups_as_keys

    groups_as_residue_idxs : list of len ng
        Contains ng arrays with the concatenated
        and sorted residues in each group
    groups_as_keys : list of len ng
        Contains ng lists with the keys
        of :obj:`frag_defs_dict` in each of groups
    """

    groups_as_keys = []
    groups_as_residue_idxs = []
    _answers = [None]*ng
    if answers is not None:
        for ii, ians in enumerate(answers):
            if isinstance(ians,str):
                _answers[ii]=ians
    answers = _answers

    if verbose:
        for key, val in frag_defs_dict.items():
            print("%s: %u-%u"%(key, val[0],val[-1]))
    for ii in range(1, ng + 1):
        print("group %u: " % ii, end='')
        if answers[ii-1] is None:
            answer = input(
                "Input a list of comma-separated posix-expressions.\n"
                "Prepend with '-' to exclude, e.g. 'TM*,-TM2,H8' to grab all TMs and H8, but exclude TM2)\n").replace(
                " ", "").strip("'").strip('"')
        else:
            answer = answers[ii-1]
        igroup, res_idxs_in_group = _mdcu.str_and_dict.match_dict_by_patterns(answer, frag_defs_dict)
        groups_as_keys.append([ilab for ilab in frag_defs_dict.keys() if ilab in igroup])
        groups_as_residue_idxs.append(sorted(res_idxs_in_group))
        print(', '.join(groups_as_keys[-1]))

    return groups_as_residue_idxs, groups_as_keys

def splice_orphan_fragments(fragments, fragnames, highest_res_idx=None,
                            orphan_name="?",
                            other_fragments=None):
    r"""
    Return a fragment list where residues not present in :obj:`fragments` are
    now new interstitial ('orphan') fragments.

    "not-present" means outside of the ranges of each fragment, s.t.
    an existing fragment like [0,1,5,6] is actually considered [0,1,2,3,4,5,6]

    Parameters
    ----------
    fragments : list
        The initial fragments potentially missing some residues
    fragnames : list
        The names of :obj:`fragments`
    highest_res_idx : int, default is None
        The highest possible res_idx. The returned fragments
        will have this idx as last idx of the last fragment (input or orphan)
        If None, the highest value (max) of fragments is
        used
    orphan_name : str, default is ?
        If the str contains a '%' character
        it will be used as a format identifier
        to use as orphan_name%ii
    other_fragments : dict, default is None
        A second set of fragment-definitions.
        If these other fragments are contained
        in the newly found orphans, then the
        orphans are re-shaped and renamed
        using this info. Typical usecase
        is for :obj:`fragments` to be
        consensus fragments (that don't
        necessarily cover the whole topology)
        and :obj:`other_fragments` to
        come from :obj:`fragments.get_fragments`
        and cover the whole topology


    Returns
    -------
    new_fragments : list
        The new fragments including those missing in the input
    new_names : list
        The new names, where the "orphans" have gotten the name orphan_char

    """
    assert len(fragments)==len(fragnames)
    if highest_res_idx is None:
        highest_res_idx = _np.max(_np.hstack(fragments))

    full_frags = []
    for ifrag, iname in zip(fragments,fragnames):
        if len(ifrag)==ifrag[-1]-ifrag[0]:
            print("Fragment %s has holes in it but will be considered from %u to %u regardless"%(iname,ifrag[0],ifrag[-1]))
        full_frags.append(_np.arange(ifrag[0],ifrag[-1]+1))
    orphans = _np.delete(_np.arange(highest_res_idx + 1), _np.hstack(full_frags))
    if len(orphans)>0:
        orphans = _get_fragments_by_jumps_in_sequence(orphans)[1]
        if other_fragments is not None:
            # The orphans have to be split using the limits of the other_fragments
            orphans = _break_fragments([frag[0] for frag in other_fragments.values()], orphans)
        if '%' in orphan_name:
            orphans_labels = [orphan_name%ii for ii, __ in enumerate(orphans)]
        else:
            orphans_labels = [orphan_name for __ in orphans]
        # The idea is that now orphans could be supersets of existing
        # fragments, s.t.
        if other_fragments is not None:
            popped = []
            for ii in range(len(orphans)):
                for xname,xfrag in other_fragments.items():
                    if xname not in popped and set(xfrag).issubset(orphans[ii]):
                        orphans[ii] = sorted(set(orphans[ii]).difference(xfrag))
                        popped.append(xname)
                        orphans.append(xfrag)
                        orphans_labels.append(xname)
        still_orphans = [ii for ii, oo in enumerate(orphans) if len(oo)>0]
        new_frags = [orphans[oo] for oo in still_orphans] + full_frags
        new_labels =[orphans_labels[oo] for oo in still_orphans] + fragnames
        idxs = _np.argsort([ifrag[0] for ifrag in new_frags])
        new_frags, new_names = [list(new_frags[ii]) for ii in idxs], [new_labels[ii] for ii in idxs]

        return new_frags, new_names
    else:
        return fragments, fragnames

def mix_fragments(highest_res_idx, consensus_frags, fragments, fragment_names):
    r"""Mix consensus frags with user-provided fragment definitions

    Wrapper around :obj:`mdciao.fragments.splice_orphan_fragments`,
    will possibly be merged with it in the future.
    The wrapper does several pre and post processing things.
    pre-processing:
     * makes the `consensus_frags` the main frags
     and the `fragments` the `other_fragments`.
     * It allows for `fragments` and/or `fragment_names`
     to be None and creates names on the fly if needed
    post-processing:
     * renames orphan fragments as sub-fragments of original
      fragments if possible
    In theory, if `fragments` covers all residues, no
    orphan's should be left

    Note
    ----
    This is an internal ad-hoc method for
    developing the best integration of
    the logic behind the fragmentation
    for flareplots, and might disappear in
    in the future. Not intended for API
    use

    Parameters:
    ----------
    highest_res_idx : int
        Typically top.n_residues - 1
    consensus_frags : dict
    fragments : list or None
        User provided fragments
    fragment_names : list or None
        User provided fragment names

    Returns:
    -------
    new_frags : list
        The new fragments
    new_names : list
        The new names
    """

    if fragments is not None:
        if fragment_names is not None:
            assert len(fragments)==len(fragment_names)
            other_frags = {fn: fr for fn, fr in zip(fragment_names, fragments)}
        else:
            other_frags = {"frag %u" % ii: fn for ii, fn in enumerate(fragments)}
    else:
        other_frags = None

    new_frags, new_names = splice_orphan_fragments(list(consensus_frags.values()),
                                          list(consensus_frags.keys()),
                                          highest_res_idx=highest_res_idx,
                                          other_fragments=other_frags,
                                          orphan_name="orphan %u")
    # todo put this into splice_orphan_fragments if it makes sense
    # reassign the orphans as sub-fragments of the parents, in case there's parents
    orphan_idxs = [ii for ii, nn in enumerate(new_names) if nn.startswith("orphan ")]
    if len(orphan_idxs) > 0 and other_frags is not None:
        parent_names = list(other_frags.keys())
        __, child = _mdcu.lists.find_parent_list([new_frags[oo] for oo in orphan_idxs],
                                                 list(other_frags.values()))
        for parent_idx, list_of_children in child.items():
            for idx, oo in enumerate(list_of_children):
                new_names[orphan_idxs[oo]] = "subfrag %u of %s"%(idx, parent_names[parent_idx])

    return new_frags, new_names


def flarekwargs_preparer(fragments, fragment_names, kwargs_freqs2flare, fixed_color_list, to_intersect_with):
    r"""
    Prepare the kwargs dictionary to call :obj:`mdciao.flare.freqs2flare` from a :obj:`mdciao.contacts.ContactGroup`


    Note
    ----
    This is a helper method that might disappear or get refactored somewhere.
    The motivation for this method is to keep freqs2flare as agnostic
    as possible with respect to things like contact-groups, interfaces,
    consensus labelers, sub-domains etc, so that the method's signature
    (already pretty long) is clear in cases where none of the above is
    needed and can be generally used to plot any set of values that have
    pair-relations associated with them.

    However, this decision somehow blurries the logic behind "fragment and residue selection"
    spreading it across two methods, because there's a partial selection
    before calling freqs2flare and a further selection inside freqs2flare.
    The upside (apart from the above mentioned conservation of generality) is
    that it allows for consistency in fragment colors and names when
    repeatedly calling the plot_freqs_as_flareplot when changing the 'scheme'
    parameter.

    Parameters
    ----------
    fragments : None or list of lists
    fragment_names : None or list of strings
    kwargs_freqs2flare : dict
    fixed_color_list : list of matplotlib colors
    to_intersect_with : the group of residues that need to be present

    Returns
    -------
    kwargs_freqs2flare : dict
    good_frags : list
    """
    good_frags = []
    if fragments is not None:
        good_frags = [ii for ii, fr in enumerate(fragments) if
                      len(_np.intersect1d(fr, to_intersect_with)) > 0]
        kwargs_freqs2flare["fragments"] = [fragments[ii] for ii in good_frags]
        kwargs_freqs2flare["colors"] = fixed_color_list[good_frags]
        if fragment_names is not None:
            assert len(fragment_names) <= len(fragment_names), \
                ValueError("Less fragment names (%u) than fragments (%u)?" % (len(fragment_names), len(fragment_names)))
            kwargs_freqs2flare["fragment_names"] = [fragment_names[ii] for ii in good_frags]
    return kwargs_freqs2flare, good_frags

def assign_fragments(res_idxs, fragments, raise_on_missing=True):
    r"""Assign a parent fragment to each residue in a list of res_idxs

    Note
    ----
    Simply wraps around :obj:`~mdciao.utils.lists.in_what_N_fragments`,
    first asserting that there's no intersection between the fragments
    and then checking whether some residxs are missing

    Parameters
    ----------
    res_idxs : iterable of ints
        The residue indices
    fragments : iterable of iterable if ints
        The fragment definitions.
        It will be checked that these definitions
        don't have common residues
    raise_on_missing : bool, default is True
        Whether to raise an Exception if
        any residue can't be found
        in the :obj:`fragments`

    Returns
    -------
    frag_idxs : np.array
        Array with the indices of
        :obj:`fragments` where each
        residue of :obj:`res_isxs`
        appears
    res_idxs : np.array
        If all residues were
        found in :obj:`fragments`,
        then this is a copy of the
        input array. If some didn't
        appear anywhere, but
        :obj:`raise_on_missing` was
        False, the missing residues
        have been deleted
    """
    #TODO very related to utils.lists.find_parent_list, perhaps merge?
    _mdcu.lists.assert_no_intersection(fragments, word="fragments")
    frag_idxs = _mdcu.lists.in_what_N_fragments(res_idxs, fragments)
    orphans = _np.flatnonzero([len(par) == 0 for par in frag_idxs])
    frag_idxs = _np.squeeze(frag_idxs)
    if len(orphans)>0:
        if raise_on_missing:
            raise ValueError("residues %s don't appear in any 'fragments'. "
                             "If you're OK with this, set 'check_if_subset=False'" % (_np.array(res_idxs)[orphans]))
        else:
            res_idxs = _np.delete(res_idxs, orphans)
            frag_idxs = _np.delete(frag_idxs, orphans)

    return frag_idxs, res_idxs


__version__ = "0.0.5"
from setuptools import setup, find_packages
import os
os.environ["BEZIER_NO_EXTENSION"]="true" # Check https://github.com/dhermes/bezier/releases/tag/2020.2.3

with open("README.rst", "r", encoding="utf-8") as fh:
    long_description = fh.read()

test_deps = [
    'coverage',
    'pytest',
    'nose',
    "mock"
]

doc_deps = [
    "sphinx",
    "sphinx-argparse",
    "sphinx-rtd-theme",
    "npsphinx",
    "sphinx-copybutton"
]

setup(
    name="mdciao",
    version=__version__,
    author_email="guillermo.perez@charite.de",
    description="mdciao: Analysis of Molecular Dynamics Simulations Using Residue Neighborhoods",
    url="https://github.com/gph82/mdciao",
    project_urls={
        "docs": "http://proteinformatics.org/mdciao",
    },
    python_requires=">=3.6",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    classifiers = ["Development Status :: 4 - Beta",
                   "Programming Language :: Python :: 3",
                   "Operating System :: OS Independent",
                   "Intended Audience :: Science/Research",
                   "License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)",
                   "Topic :: Scientific/Engineering :: Chemistry",
                   "Topic :: Scientific/Engineering :: Bio-Informatics",
                   "Topic :: Scientific/Engineering :: Visualization",
                   ],
    packages=find_packages(),
    install_requires=[
                    "cython",
                    "numpy>=1.18.1",
                    "mdtraj",
                    "astunparse; python_version!='3.8'",
                    "astunparse<1.6.3; python_version=='3.8'",
                    "pandas",
                    "matplotlib",
                    "scipy",
                    "joblib",
                    "openpyxl",
                    "biopython>=1.77",
                    "ipython",
                    "XlsxWriter",
                    "requests",
                    "tqdm",
                    "natsort",
                    "bezier; python_version!='3.6'",
                    "bezier<2020.2.3; python_version=='3.6'",
                    "mpl_chord_diagram>=0.3.2"
                     ]
                     +test_deps
                     #+doc_deps
    # tests_require=test_deps, # has been deprecated, gotta learn how to use tox
    ,
    data_files=[
        ('data_for_mdciao/examples', ['tests/data/examples/gs-b2ar.noH.stride.5.xtc',
                                      'tests/data/examples/gs-b2ar.noH.pdb']),
        ('data_for_mdciao/nomenclature', ['tests/data/nomenclature/adrb2_human.xlsx',
                                          'tests/data/nomenclature/CGN_3SN6.txt',
                                          'tests/data/nomenclature/nomenclature.bib',
                                          'tests/data/nomenclature/KLIFS_P31751.xlsx']),
        ('data_for_mdciao/json', ['tests/data/json/tip.json']),
        ('data_for_mdciao/RCSB_pdb', ['tests/data/RCSB_pdb/3SN6.pdb.gz']),
        ("data_for_mdciao/notebooks", ["mdciao/examples/Tutorial.ipynb"]),
        ("data_for_mdciao/notebooks", ["mdciao/examples/Missing_Contacts.ipynb"]),
        ("data_for_mdciao/notebooks", ["mdciao/examples/Manuscript.ipynb"]),
        ("data_for_mdciao/notebooks", ["mdciao/examples/Flareplot_Schemes.ipynb"]),
        ("data_for_mdciao/notebooks", ["mdciao/examples/EGFR_Kinase_Inhibitors.ipynb"]),
        ("data_for_mdciao/notebooks", ["mdciao/examples/Comparing_CGs_Bars.ipynb"]),
        ("data_for_mdciao/notebooks", ["mdciao/examples/Comparing_CGs_Flares.ipynb"])
    ],
    scripts=['scripts/mdc_neighborhoods.py',
             'scripts/mdc_sites.py',
             'scripts/mdc_interface.py',
             'scripts/mdc_fragments.py',
             'scripts/mdc_GPCR_overview.py',
             'scripts/mdc_CGN_overview.py',
             'scripts/mdc_KLIFS_overview.py',
             'scripts/mdc_compare.py',
             'scripts/mdc_examples.py',
             'scripts/mdc_pdb.py',
             'scripts/mdc_residues.py',
             'scripts/mdc_notebooks.py'
             #'scripts/density_by_sites.py',
             #'scripts/site_figures.py',
             #'scripts/residue_dihedrals.py',
             #'scripts/contact_maps.py'
             #'scripts/CARDS.py'
             ],
)




#!/usr/bin/env python3

##############################################################################
#    This file is part of mdciao.
#    
#    Copyright 2020 Charité Universitätsmedizin Berlin and the Authors
#
#    Authors: Guillermo Pérez-Hernandez
#    Contributors:
#
#    mdciao is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    mdciao is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with mdciao.  If not, see <https://www.gnu.org/licenses/>.
##############################################################################


#TODO i'm not putting notebooks in cli to avoid cicular dependencies (ContactGroupL394 imports from cli and is in examples)
from mdciao.examples import notebooks as notebooks
from mdciao.parsers import parser_for_notebooks
parser = parser_for_notebooks()
a = parser.parse_args()
notebooks()


##############################################################################
#    This file is part of mdciao.
#    
#    Copyright 2020 Charité Universitätsmedizin Berlin and the Authors
#
#    Authors: Guillermo Pérez-Hernandez
#    Contributors:
#
#    mdciao is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    mdciao is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with mdciao.  If not, see <https://www.gnu.org/licenses/>.
##############################################################################

import numpy as _np
import mdtraj as _md
import matplotlib.pyplot as _plt
from matplotlib import rcParams as _rcParams

def xtcs2dihs(xtcs, top, dih_idxs, stride=1, consolidate=True,
              chunksize=1000, return_time=False,
              n_jobs=1,
              progressbar=False,
              ):
    """Returns the time-dependent traces of sidechain angles

    Parameters
    ----------
    xtcs : list of strings
        list of filenames with trajectory data. Typically xtcs,
        but can be any type of file readable by :obj:mdtraj
    top : str or :py:class:`mdtraj.Topology`
        Topology that matches :obj:xtcs
    dih_idxs : iterable
        List of quadroplets with atom idxs (zero-indexed) of the dihedrals
    stride : int, default is 1
        Stride the trajectory data down by this value
    consolidate : boolean, default is True
        Return the time-traces consolidated
        into one array by using np.vstack
    chunksize : integer, default is 1000
        How many frames will be read into memory for
        computation of the contact time-traces. The higher the number,
        the higher the memory requirements
    n_jobs : int, default is 1
        to how many processors to parallellize

    return_time : boolean, default is False
        Return also the time array in ps

    Returns
    -------
    chis, or
    chis, time_trajs if return_time=True

    """

    from tqdm import tqdm

    if progressbar:
        iterfunct = lambda a : tqdm(a)
    else:
        iterfunct = lambda a : a
    ictcs_itimes = _Parallel(n_jobs=n_jobs)(_delayed(per_xtc_dihs)(top, ixtc, dih_idxs,chunksize,stride,ii,
                                                                  )
                                            for ii, ixtc in enumerate(iterfunct(xtcs)))
    dihs = []
    times = []
    for idihs, itimes in ictcs_itimes:
        dihs.append(idihs)
        times.append(itimes)

    if consolidate:
        try:
            adihs = _np.vstack(dihs)
            times = _np.hstack(times)
        except ValueError as e:
            print(e)
            print([_np.shape(ic) for ic in dihs])
            raise
    else:
        adihs = dihs
        times = times

    if not return_time:
        return adihs
    else:
        return adihs, times

def per_xtc_dihs(top, ixtc, dih_idxs, chunksize, stride,
                traj_idx,
                **mddih_kwargs):

    iterate, inform = iterate_and_inform_lambdas(ixtc,stride, chunksize, top=top)
    idihs = []
    running_f = 0
    inform(ixtc, traj_idx, 0, running_f)
    itime = []
    for jj, igeom in enumerate(iterate(ixtc)):
        running_f += igeom.n_frames
        inform(ixtc, traj_idx, jj, running_f)
        itime.append(igeom.time)
        jdihs = _md.compute_dihedrals(igeom, dih_idxs, **mddih_kwargs)


        idihs.append(jdihs)

    itime = _np.hstack(itime)
    idihs = _np.vstack(idihs)

    return idihs, itime

def plot_dih(ictc, iax,
                 color_scheme=None,
                 ctc_cutoff_Ang=0,
                 n_smooth_hw=0,
                 dt=1,
                 gray_background=False,
                 shorten_AAs=False,
                 t_unit='ps',
                 max_handles_per_row=4,
               ylabel='dih / deg',
             lambda_ang=None
                 ):
    if color_scheme is None:
        color_scheme = _rcParams['axes.prop_cycle'].by_key()["color"]
    if lambda_ang is None:
        lambda_ang = lambda x : x
    color_scheme = _np.tile(color_scheme, _np.ceil(ictc.n_trajs/len(color_scheme)).astype(int)+1)
    iax.set_ylabel(ylabel, rotation=90)

    for traj_idx, (ictc_traj, itime, trjlabel) in enumerate(zip(ictc.feat_trajs,
                                                                ictc.time_trajs,
                                                                ictc.trajstrs)):

        ilabel = '%s'%trjlabel

        #print(ictc_traj*input2Ang)
        plot_w_smoothing_auto(iax, itime * dt, lambda_ang(ictc_traj),
                              ilabel,
                              color_scheme[traj_idx],
                              gray_background=gray_background,
                              n_smooth_hw=n_smooth_hw)
    iax.legend(loc=1, fontsize=_rcParams["font.size"]*.75,
               ncol=_np.ceil(ictc.n_trajs/max_handles_per_row).astype(int)
               )
    ctc_label = ictc.label
    if shorten_AAs:
        ctc_label = ictc.ctc_label_short
    ctc_label = ctc_label.replace("@None","")

    iax.text(_np.mean(iax.get_xlim()),
             iax.get_ylim()[0]+_np.abs(_np.diff(iax.get_ylim()))*.1,
             ctc_label,
             ha='center')

    iax.set_xlabel('t / %s' % _replace4latex(t_unit))
    iax.set_xlim([0, ictc.time_max * dt])
    #iax.set_ylim([0,iax.get_ylim()[1]])

class angle(object):
    r"""Class for storing everything related to a contact"""
    #todo consider packing some of this stuff in the site_obj class
    def __init__(self, atom_idx_quad,
                 dih_trajs,
                 time_arrays,
                 top=None,
                 trajs=None,
                 res_idx=None,
                 ang_type=None,
                 fragment_idx=None,
                 fragment_name=None,
                 fragment_color=None,
                 consensus_label=None):
        """

        Parameters
        ----------
        atom_idx_quad : iterable for 4 atom indices for this dihedral.
        dih_trajs : list of list, the code converts it into a list of array
        time_arrays : list of list,
        top : :py:class:`mdtraj.Topology`
        trajs:
        fragment_idx :
        fragment_name :
        fragment_color :
        anchor_residue_idx :
        consensus_label :
        """

        self.atom_idx_quad = atom_idx_quad
        self._res_idx = res_idx
        self._ang_type = ang_type
        self._dih_trajs = [_np.array(itraj) for itraj in dih_trajs]
        self._top = top
        self._trajs = trajs

        self._time_arrays = time_arrays
        self._n_trajs = len(dih_trajs)
        assert self._n_trajs == len(time_arrays)
        assert all([len(itraj) == len(itime) for itraj, itime in zip(dih_trajs, time_arrays)])
        self._time_max = _np.max(_np.hstack(time_arrays))

        self._consensus_label = consensus_label
        self._fragment_idx  = fragment_idx
        if fragment_name is None:
            # assert self.idxs is not None
            # self._fragment_names = self._fragment_idxs

            if self.fragment_idx is not None:
                self._fragment_name = self._fragment_idx
            else:
                self._fragment_name = None
        else:
            self._fragment_name = fragment_name
        self._fragment_color = fragment_color

    @property
    def fragment_name(self):
        """

        Returns
        -------
        list of list, Fragment names if passed, else fragment idxs. If both are not available then None(default)

        """
        return self._fragment_name

    @property
    def fragment_idx(self):
        """

        Returns
        -------
        list of list, Fragment idxs if passed, else None(default)

        """
        return self._fragment_idx

    @property
    def time_max(self):
        """

        Returns
        -------
        int or float, maximum time from list of list of time

        """
        return self._time_max

    @property
    def trajlabels(self):
        """

        Returns
        -------
        list, list of labels for each trajectory if passed.
        If labels are not passed then labels like 'traj 0','traj 1' and so on are assigned

        """
        if self.trajs is None:
            return ['traj %u'%ii for ii in range(self.n_trajs)]
        else:
            return self.trajs

    @property
    def label(self):
        return self.dih_label_short_latex

    @property
    def n_trajs(self):
        """

        Returns
        -------
        int, total number of trajectories that were passed.

        """
        return self._n_trajs

    @property
    def n_frames(self):
        """

        Returns
        -------
        list, list of frames in each trajectory.

        """
        return [len(itraj) for itraj in self.dih_trajs]



    @property
    def res_idx(self):
        """

        Returns
        -------
        list of residue index pair passed

        """
        return self._res_idx

    @property
    def residue_name(self):
        """

        Returns
        -------
        list, for each residue index in the residue contact pair, the corresponding residue name from the topology file.
        example : ['GLU30','VAL212']

        """
        if self.res_idx is not None:
            return str(self.topology.residue(self.res_idx))
        else:
            return None

    @property
    def residue_name_short(self):
        """

        Returns
        -------
        list, for each residue name in the residue contact pair, the corresponding short residue name from the topology file.
        example : ['E30', 'V212']

        """
        from .residue_and_atom_utils import shorten_AA as _shorten_AA
        if self.res_idx is not None:
            return _shorten_AA(self.residue_name, substitute_fail="long", keep_index=True)
        else:
            return None

    @property
    def feat_trajs(self):
        return self.dih_trajs

    @property
    def ang_type(self):
        return self._ang_type

    @property
    def dih_label(self):
        """

        Returns
        -------
        str,

        """
        dih_label = '[%s]'%('-'.join([str(self.top.atom(ii)) for ii in self.atom_idx_quad]))

        if self.ang_type is not None:
            dih_label = '%s %s'%(self.ang_type,dih_label)

        return dih_label


    @property
    def dih_label_short(self):
        """

        Returns
        -------
        str,

        """

        if self.ang_type is not None:
            return self.ang_type
        else:
            return self.dih_label



    @property
    def time_arrays(self):
        """

        Returns
        -------

        """
        return self._time_arrays

    @property
    def dih_trajs(self):
        """

        Returns
        -------

        """
        return self._dih_trajs

    @property
    def trajs(self):
        """

        Returns
        -------

        """
        return self._trajs

    @property
    def fragment_color(self):
        """

        Returns
        -------

        """
        return self._fragment_color

    @property
    def top(self):
        """

        Returns
        -------

        """
        return self._top

    @property
    def topology(self):
        """

        Returns
        -------

        """
        return self._top

    @property
    def consensus_label(self):
        """

        Returns
        -------

        """
        return self._consensus_label

    @property
    def dih_label_short_latex(self):
        if self.ang_type is not None:
            if self.ang_type.lower().startswith("chi"):
                return '$\\chi_{%s}$'%(self.ang_type[-1])
            else:
                return '$\\%s$'%(self.ang_type)
        else:
            return None

    def __str__(self):
        out = "Angle object for angle defined by atoms"
        out += '\n%s'%self.atom_idx_quad
        #out += "\n%s"%self.res_idx
        out += "\nFor %u trajectories"%self.n_trajs
        for var in dir(self):
            if not var.startswith("_"):
                out += '\n%s: %s'%(var, getattr(self,'%s'%var))
        return out

def plot_ramachandran(phi, psi, iax=None, panelsize=5, bins=72):
        if iax is None:
            _plt.figure(figsize=(panelsize,panelsize))
            iax=_plt.gca()
            h, x, y = _np.histogram2d(_np.hstack((phi,psi)))

def residue_dihedrals(topology, trajectories, resSeq_idxs,
                      res_idxs=False,
                      stride=1,
                      chunksize_in_frames=10000,
                      n_smooth_hw=0,
                      ask=True,
                      sort=True,
                      fragmentify=True,
                      fragment_names="",
                      graphic_ext=".pdf",
                      output_ascii=None,
                      BW_uniprot="None",
                      CGN_PDB="None",
                      output_dir='.',
                      color_by_fragment=True,
                      output_desc='dih',
                      t_unit='ns',
                      curve_color="auto",
                      gray_background=False,
                      graphic_dpi=150,
                      short_AA_names=False,
                      write_to_disk_BW=False,
                      plot_timedep=True,
                      n_cols=4,
                      types='all',
                      n_jobs=1,
                      use_deg=True,
                      use_cos=False,
                      ):
    ang2plotfac = 1
    xlabel, ang_u = 'dih', 'rad'
    bins = 72
    ang_lambda = lambda x : x
    xlim =_np.array([-_np.pi, +_np.pi])
    if use_deg:
        ang_lambda = lambda x : (180 / _np.pi) * x
        ang_u = 'deg'
        xlim = ang_lambda(xlim)
    if use_cos:
        ang_lambda = lambda x : _np.cos(x)
        ang_u = 'arb. u.'
        xlabel = '$cos(dih)$'
        bins = 100
        xlim = [-1,1]

    if resSeq_idxs is None:
        print("You have to provide some residue indices via the --resSeq_idxs option")
        return None

    dt = _t_unit2dt(t_unit)

    _offer_to_create_dir(output_dir)

    # String comparison to allow for command line argparse-use directly
    if str(output_ascii).lower() != 'none' and str(output_ascii).lower().strip(".") in ["dat", "txt", "xlsx"]:
        ascii_ext = str(output_ascii).lower().strip(".")
    else:
        ascii_ext = None


    _resSeq_idxs = rangeexpand(resSeq_idxs)
    if len(_resSeq_idxs) == 0:
        raise ValueError("Please check your input indices, "
                         "they do not make sense %s" % resSeq_idxs)
    else:
        resSeq_idxs = _resSeq_idxs
    if sort:
        resSeq_idxs = sorted(resSeq_idxs)

    xtcs = sorted(trajectories)
    try:
        print("Will compute contact frequencies for the files:\n"
              "%s\n with a stride of %u frames.\n" % (
            "\n  ".join(xtcs), stride))
    except:
        pass

    if isinstance(topology, str):
        refgeom = _md.load(topology)
    else:
        refgeom = topology
    if fragmentify:
        fragments = get_fragments(refgeom.top,method='bonds')
    else:
        raise NotImplementedError("This feature is not yet implemented")

    fragment_names = _parse_fragment_naming_options(fragment_names, fragments, refgeom.top)

    # Do we want BW definitions
    BW = _parse_consensus_option(BW_uniprot, 'BW', refgeom.top, fragments, write_to_disk=write_to_disk_BW)

    # Dow we want CGN definitions:
    CGN = _parse_consensus_option(CGN_PDB, 'CGN', refgeom.top, fragments)

    # TODO find a consistent way for coloring fragments
    fragcolors = [cc for cc in mycolors]
    fragcolors.extend(fragcolors)
    fragcolors.extend(fragcolors)
    if isinstance(color_by_fragment, bool) and not color_by_fragment:
        fragcolors = ['blue' for cc in fragcolors]
    elif isinstance(color_by_fragment, str):
        fragcolors = [color_by_fragment for cc in fragcolors]

    if res_idxs:
        resSeq2residxs = {refgeom.top.residue(ii).resSeq: ii for ii in resSeq_idxs}
        print("\nInterpreting input indices as zero-indexed residue indexes")
    else:
        resSeq2residxs, _ = _interactive_fragment_picker_by_resSeq(resSeq_idxs, fragments, refgeom.top,
                                                                  pick_first_fragment_by_default=not ask,
                                                                  additional_naming_dicts={"BW":BW,"CGN":CGN}
                                                                  )

    print('%10s  %10s  %10s  %10s %10s %10s' % tuple(("residue  residx fragment  resSeq BW  CGN".split())))
    for key, val in resSeq2residxs.items():
        print('%10s  %10u  %10u %10u %10s %10s' % (refgeom.top.residue(val), val, in_what_fragment(val, fragments),
                                                  key,
                                                  BW[val], CGN[val]))

    quad_dict_by_res_idxs = _dih_idxs_for_residue(resSeq2residxs.values(),refgeom)
    if types.lower()=='backbone':
        quad_dict_by_res_idxs={key:{key2:val2 for key2, val2 in val.items() if key2.lower() in ["phi","psi"]} for key, val in quad_dict_by_res_idxs.items()}
    elif types.lower()=='sidechain':
        quad_dict_by_res_idxs = {key: {key2: val2 for key2, val2 in val.items() if key2.lower() not in ["phi", "psi"]} for
                                 key, val in quad_dict_by_res_idxs.items()}
    elif types.lower()=='all':
        pass
    else:
        raise ValueError(types)
    dih_idxs =_np.vstack([[val2 for val2 in val.values()] for val in quad_dict_by_res_idxs.values()])
    dih_trajs, time_array = xtcs2dihs(xtcs, refgeom.top, dih_idxs , stride=stride,
                                       chunksize=chunksize_in_frames, return_time=True,
                                       consolidate=False,
                                       n_jobs=n_jobs,

                                       )
    print()
    # Create per-residue dicts with angle objects
    angles = {}
    idx_iter = iter(range(len(dih_idxs)))
    for res_idx, quad_dict in quad_dict_by_res_idxs.items():
        angles[res_idx] = []
        for ang_type, iquad in quad_dict.items():
            consensus_label = _choose_between_consensus_dicts(res_idx, [BW, CGN])
            fragment_idx =    in_what_fragment(res_idx, fragments)
            idx = next(idx_iter)
            angles[res_idx].append(angle(iquad,
                                         [itraj[:, idx] for itraj in dih_trajs],
                                         time_array,
                                        res_idx=res_idx,
                                         ang_type=ang_type,
                                         top=refgeom.top,
                                         consensus_label=consensus_label,
                                         trajs=xtcs,
                                         fragment_idx=fragment_idx,
                                         fragment_name=fragment_names[fragment_idx],
                                         fragment_color=fragcolors[fragment_idx],
                                         ))

    panelheight = 3
    n_cols =_np.min((n_cols, len(resSeq2residxs)))
    n_rows =_np.ceil(len(resSeq2residxs) / n_cols).astype(int)
    panelsize = 4
    panelsize2font = 3.5
    histofig, histoax = _plt.subplots(n_rows, n_cols,
                                      sharex=True,
                                      sharey=True,
                                      figsize=(n_cols * panelsize * 2, n_rows * panelsize), squeeze=False)

    # One loop for the histograms
    _rcParams["font.size"]=panelsize*panelsize2font
    for jax, res_idx in zip(histoax.flatten(),
                            angles.keys()):
        for idih in angles[res_idx]:
            h, x = _np.histogram(ang_lambda(_np.hstack(idih.dih_trajs)),
                                 bins=bins)
            jax.plot(x[:-1]*ang2plotfac,h,label=idih.dih_label_short_latex)
            jax.legend()
        jax.set_title(idih.residue_name)
    jax.set_xlim(xlim)
    [jax.set_xlabel('%s / %s'%(xlabel, ang_u)) for jax in histoax[-1]]
    histofig.tight_layout(h_pad=2, w_pad=0, pad=0)
    fname = "%s.overall.%s" % (output_desc.strip("."), graphic_ext.strip("."))
    fname = _path.join(output_dir, fname)
    histofig.savefig(fname, dpi=graphic_dpi)
    print("The following files have been created:")
    print(fname)

    # One loop for the time resolved neighborhoods
    if plot_timedep:
        for res_idx, res_dihs in angles.items():
            fname = '%s.%s.time_resolved.%s' % (output_desc.strip("."),
                                                idih.residue_name,
                                                graphic_ext.strip("."))
            fname = _path.join(output_dir, fname)
            n_rows=len(res_dihs)
            myfig, myax = _plt.subplots(n_rows, 1,
                                        figsize=(10, n_rows * panelheight),
                                        squeeze=False)
            myax = myax[:, 0]
            for iax, idih in zip(myax, res_dihs):

                # Plot individual angles
                from .contacts import plot_dih
                plot_dih(idih, iax,
                         color_scheme=_my_color_schemes(curve_color),
                         n_smooth_hw=n_smooth_hw,
                         dt=dt,
                         t_unit=t_unit,
                         gray_background=gray_background,
                         #shorten_AAs=short_AA_names,
                         lambda_ang=ang_lambda,
                             )

            # One title for all axes on top
            title = idih.residue_name
            if short_AA_names:
                title = idih.residue_name
            myfig.axes[0].set_title(title)

            myfig.tight_layout(h_pad=0,w_pad=0,pad=0)
            myfig.savefig(fname, bbox_inches="tight", dpi=graphic_dpi)
            _plt.close(myfig)
            print(fname)
            #ihood.save_trajs(output_desc,ascii_ext,output_dir, dt=dt,t_unit=t_unit, verbose=True)
            print()

    return None

def _get_dih_idxs(geom):
    methods = {
        "phi" : 'compute_phi',
        "psi":  'compute_psi',
        "chi1": 'compute_chi1',
        "chi2": 'compute_chi2',
        "chi3": 'compute_chi3',
        "chi4": 'compute_chi4',
        "chi5": 'compute_chi5',
    }
    idxs = {}
    for key, imeth in methods.items():
        idxs[key] = getattr(_md, imeth)(geom[0])[0]
    return idxs

def _angle_quadruplet2residx(quadruplets, top, max_residues_per_quad=1):
    r"""

    :param quadruplets:
    :param top:
    :param max_residues_per_quad:
    :return:
    """
    allow_only_complimentary = 4 - _np.arange(max_residues_per_quad)
    rowidx2residx = []
    for row in quadruplets:
        res_row = _np.array([top.atom(ii).residue.index for ii in row])

        # Because the bincount is done on integers representing the
        # true residue index as they appears in the topology,
        # these idxs are true idxs and robust towards omission
        # (e.g. the residue in question did not have computable angles)
        bc = _np.bincount(res_row)
        if _np.max(bc) in allow_only_complimentary:
            rowidx2residx.append(_np.argmax(bc))
        else:
            raise ValueError("Cannot assign quad %s (res_idxs %s) when only %u residues are allowed per quad"%(row,res_row,max_residues_per_quad))
            #rowidx2residx.append(None)

    return _np.array(rowidx2residx)

def _dih_idxs_for_residue(res_idxs, geom):
    quads = _get_dih_idxs(geom)
    quad_idxs_2_res_idxs = {key:_angle_quadruplet2residx(val, geom.top, max_residues_per_quad=2) for key, val in quads.items()}
    dict_out = {}
    for ii in res_idxs:
        quad_idx = {key:_np.argwhere(val==ii).squeeze() for key, val in quad_idxs_2_res_idxs.items()}
        dict_out[ii] = {key:quads[key][val] for key, val in quad_idx.items() if _np.size(val)>0}
    return dict_out




r"""
Functions related to Center-Of-Mass (COM) computations

.. autosummary::
   :nosignatures:
   :toctree: generated/


"""
import numpy as _np
def geom2COMdist(igeom, residue_pairs):
    r"""
    Returns the distances between center-of-mass (COM)
    Parameters
    ----------
    igeom: :obj:`mdtraj.Trajectory`
    residue_pairs: iterable of integer pairs
        pairs of residues by their zero-indexed serial indexes

    Returns
    -------
    COMs_array : np.ndarray of shape(igeom.n_frames, len(res_pairs))
        contains the time-traces of the residue COMs
    """
    from scipy.spatial.distance import pdist, squareform
    residue_idxs_unique, pair_map = _np.unique(residue_pairs, return_inverse=True)
    pair_map = pair_map.reshape(len(residue_pairs),2)


    COMs_xyz = geom2COMxyz(igeom, residue_idxs=residue_idxs_unique)[:,residue_idxs_unique]

    # Only do pdist of the needed residues
    COMs_dist_triu = _np.array([pdist(ixyz) for ixyz in COMs_xyz])

    # Get square matrices
    COMs_square = _np.array([squareform(trupper) for trupper in COMs_dist_triu])

    # Use the pair map to get the right indices
    COMs_array = _np.array([COMs_square[:,ii,jj] for ii,jj in pair_map]).T

    return COMs_array

def geom2COMxyz(igeom, residue_idxs=None):
    r"""
    Returns the time trace of per-residue
    center-of-masses (COMs) in cartesian coordinates

    Parameters
    ----------
    igeom : :obj:`mdtraj.Trajectory`

    residue_idxs : iterable, default is None
        Residues for which the center of mass will be computed. Default
        is to compute all residues. The excluded residues will appear
        as np.nans in the returned value, which has the same shape
        regardless of the input

    Returns
    -------
    COMs : numpy.ndarray of shape (igeom.n_frames, igeom.n_residues,3)

    """

    #TODO check the behaviour of COMs always having the same shape
    if residue_idxs is None:
        residue_idxs=_np.arange(igeom.top.n_residues)
    masses = [_np.hstack([aa.element.mass for aa in rr.atoms]) for rr in
              igeom.top.residues]
    COMs_res_time_coords = _np.zeros((igeom.n_residues,igeom.n_frames,3))
    COMs_res_time_coords[:,:,:] = _np.nan
    COMs_res_time_coords[residue_idxs] = [_np.average(igeom.xyz[:, [aa.index for aa in igeom.top.residue(index).atoms], :], axis=1, weights=masses[index])
                            for index in residue_idxs]
    COMs_time_res_coords = _np.swapaxes(_np.array(COMs_res_time_coords),0,1)
    return COMs_time_res_coords


##############################################################################
#    This file is part of mdciao.
#    
#    Copyright 2020 Charité Universitätsmedizin Berlin and the Authors
#
#    Authors: Guillermo Pérez-Hernandez
#    Contributors:
#
#    mdciao is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    mdciao is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with mdciao.  If not, see <https://www.gnu.org/licenses/>.
##############################################################################

r"""
These methods attempt to compute the overlaps
between Text objects in a matplotlib plot. These methods
are used by mdciao.flare exclusively, trying to
work around potential re-sizing problems by
trying to avoid re-drawings of the axis (and some shameless
fudging). Naturally, this can be error prone and hard to test.

That's why, for the moment the only tests are that
1) all code is covered by normal execution and nothing goes wrong and
b) visual inspection of the notebooks testing the flareplot schemes

In the future, I'll either understand better how matplotlib
bounding boxes work and/or use libs like shapely to detect
overlaps and stuff.
"""

import numpy as _np

def outermost_text_corner(texts,  center=(0,0), verbose=False,):
    r"""
    Return the distance of the farthest corner of any Text's FancyBoxPatch

    Will fail if:
     * axes.draw() hasn't been called on this axis before
     * text wasn't instantiated with the bbox argument
       Check https://matplotlib.org/stable/api/text_api.html#matplotlib.text.Text.get_bbox_patch
       for more info on FancyBoxPatch

    Parameters
    ----------
    texts :  list
        :obj:`~matplotlib.text.Text` objects
    center : tuple, default is (0,0)
    verbose : bool, default is False

    Returns
    -------
    d : float

    """
    rmax = 0
    for ii, t in enumerate(texts):
        verts = text2FBPverts(t)
        ir = _np.sqrt(_np.sum((verts-_np.array(center)) ** 2, axis=1)).max()
        if ir>rmax:
            rmax = ir
            idx = ii
    if verbose:
        print("Outermost textbox has index %u and is "%idx, texts[idx])
    return rmax


def plot_fancypatches(texts, lw=5, verbose=False, renderer=None):
    iax = texts[0].axes
    iax.draw(renderer or iax.figure.canvas.get_renderer())
    for t in texts:
        verts = text2FBPverts(t)
        iax.plot(*verts.T, ls="-", lw=lw)
        if verbose:
            print(t)
            print(verts)
            print()

def any_overlap_via_FancyBoxPach(text_objects_1, text_objects_2):
    r"""
    Return the first pair of text objects that overlap

    The texts's FancyBoxPatch is used to check for overlap, since
    it's rotated with the text's orientation. Check
    https://matplotlib.org/stable/api/text_api.html#matplotlib.text.Text.get_bbox_patch
    for more info on FancyBoxPatch

    The self-overlap is ignored if the lists share members

    Parameters
    ----------
    text_objects_1 : list
        List of :obj:`~matplotlib.text.Text` objects

    text_objects_1 : list
        List of :obj:`~matplotlib.text.Text` objects
    Returns
    -------
    pair : list of two text objects or None
    """

    boxes_2 = {}
    edges_2 = {}
    iax = text_objects_1[0].axes
    renderer = iax.figure.canvas.get_renderer()
    #iax.draw(renderer) # otherwise the Fancyboxes get wrong sizes
    for t1 in text_objects_1:
        t1.draw(renderer)
        edges_1 = text2FBPedges(t1)
        for idx2, t2 in enumerate(text_objects_2):
            if t1 is not t2:
                if idx2 not in boxes_2.keys():
                    t2.draw(renderer)
                    edges_2[idx2] =  text2FBPedges(t2)
                    # draw_line(l2,iax,"b",lw=5)
                    xy = FBintersect_via_edges(edges_1, edges_2[idx2])
                    if xy is not False:
                        # print(l2)
                        #draw_line(l1,iax,"g",lw=5)
                        #t1.axes.plot(xy[0], xy[1], "or", ms=10)
                        #plot_fancypatches([t1, t2], lw=.5)
                        return [t1, t2]

def FBintersect_via_edges(rect1, rect2):
    r"""
    Whether to rectangles intersect

    Rectangles are defined as the edges connecting
    their four corners, i.e. A-B, B-C, C-D, D-A

    Parameters
    ----------
    rect1 : _np.ndarray of shape (4,4)
        Each line of the rect is an
        edge of shape [xA, yA, xB, yB]
        where A and B are the points
        connected by the edge
    rect2 : _np.ndarray of shape (4,4)
        Each line of the rect is an
            edge of shape [xA, yA, xB, yB]
            where A and B are the points
            connected by the edge

    Returns
    -------
    x, y : Tuple or False
        The coordinates of the first intersection
        or False if there is none
    """
    for e1 in rect1:
        for e2 in rect2:
            return  segment_intersection(e1.reshape((2, 2)), e2.reshape((2, 2)))


def text2FBPedges(txt):
    r""" Return, in data units, the edges of the FancyBoxPatch of a Text object

    Parameters
    ----------
    txt : :obj:`~matplotlib.text.Text`

    Returns
    -------
    edges : 2D :obj:`numpy.ndarray` of shape (4,4)
        The edges connecting the vertices of
        the FancyBoxPatch. Each edge consists of
        four floats [xA, yA, xB, yB] where A and
        B are the vertices connected by the edge
    """
    verts_1 = text2FBPverts(txt)
    return _np.vstack([_np.hstack([l1, l2]) for l1, l2 in zip(verts_1[:-1], verts_1[1:])])

def text2FBPverts(txt):
    r"""
    Return, in data units, the vertices of the FancyBoxPatch of a Text object

    Will fail if the Text object wasn't instantiated with a bbox dictionary, because
    then it doesn't have a FancyBoxPatch.

    The advantage of the FancyBoxPatch is that it's rotated with the text
    and it can be used to avoid text-overlap better

    See the matplotlib doc for more info
    https://matplotlib.org/stable/api/text_api.html#matplotlib.text.Text

    Parameters
    ----------
    txt : :obj:`~matplotlib.text.Text`

    Returns
    -------
    verts : 2D :obj:`numpy.ndarray` of shape (5,2)
        The vertices of the FancyBoxPatch. Usually,
        the first and last one are the same, so
        that the can be used to create a closed polygonpatch
        in matplotlib
    """
    iax = txt.axes
    txt.update_bbox_position_size(renderer=iax.figure.canvas.get_renderer())
    bbox = txt.get_bbox_patch()
    return iax.transData.inverted().transform(bbox.get_verts())

def draw_line(line, iax,color="k",lw=2):
    r"""

    Parameters
    ----------
    line : x1,y1, x2, y2

    Returns
    -------

    """
    iax.plot((line[0], line[2]), (line[1], line[3]),"-o", color=color, lw=lw)

def line_intersection(line1, line2):
    r"""
    Check if two lines (not segments) intersect

    The line passes through the points (A,B) and (C,D)

    Parameters
    ----------
    line1 : iterable of pairs
        [[xA, yA],[xB, yB]]
    line2 : iterable of pairs
        [[xC, yC],[xD, yD]]

    Returns
    -------
    x, y : tuple
        coordinates of the intersection
        None when there is no intersection
    """
    xdiff = (line1[0][0] - line1[1][0], line2[0][0] - line2[1][0])
    ydiff = (line1[0][1] - line1[1][1], line2[0][1] - line2[1][1])

    def det(a, b):
        return a[0] * b[1] - a[1] * b[0]

    div = det(xdiff, ydiff)
    if div == 0:
        return None
        #raise Exception('lines do not intersect')

    d = (det(*line1), det(*line2))
    x = det(d, xdiff) / div
    y = det(d, ydiff) / div
    return x, y

def segment_intersection(line1,line2):
    r"""
    Check if two segments intersect

    The segments are the line connecting the points (A,B) and (C,D)

    Parameters
    ----------
    line1 : iterable of pairs
        [[xA, yA],[xB, yB]]
    line2 : iterable of pairs
        [[xC, yC],[xD, yD]]

    Returns
    -------
    x, y : tuple or False
        coordinates of the intersection
        False when there is no intersection or
        the intersection lies outside the segments
    """
    xy = line_intersection(line1,line2)
    if xy is not None:
        inside_segments = []
        for line in [line1, line2]:
            vec1, vec2 = xy-line[0], xy-line[1]
            inside_segments.append(_np.dot(vec1,vec2) < 0 )# vectors point in opposite directions
        if all(inside_segments):
            return xy
        else:
            return False
    else:
        return False



r"""
Miscellaneous operations on list or list-like objects
.. autosummary::
   :nosignatures:
   :toctree: generated/


"""

import numpy as _np

from itertools import groupby as _groupby
from collections import defaultdict as _defdict
def contiguous_ranges(list_in):
    r"""
    For every unique entry in :obj:`list_in` return the contiguous ranges in list

    Parameters
    ----------
    list_in : list

    Returns
    -------
    ranges : dict
        The keys are with unique entries of list_in, values are the ranges
        in which the entry appears

    """
    offset = 0
    _ranges = _defdict(list)
    for key, grpr in _groupby(list_in):
        l = len(list(grpr))
        irange = _np.arange(offset, offset + l)
        _ranges[key].append(irange)
        offset+=l
    return dict(_ranges)

def re_warp(array_in, lengths):
    """Return iterable ::py:obj:array_in as a list of arrays, each
     one with the length specified in lengths

    Parameters
    ----------

    array_in: any iterable
        Iterable to be re_warped

    lengths : int or iterable of integers
        Lengths of the individual elements of the returned array. If only one int is parsed, all lengths will
        be that int. Special cases:
            * more lengths than needed are parsed: the last elements of the returned value are empty
            until all lengths have been used
            * less lengths than array_in could take: only the lenghts specified are returned in the
            warped list, the rest is unreturned
    Returns
    -------
    warped: list
    """

    if _np.ndim(lengths)==0:
        lengths = [lengths] * int(_np.ceil(len(array_in) / lengths))

    warped = []
    idxi = 0
    for ii, ll in enumerate(lengths):
        warped.append(array_in[idxi:idxi+ll])
        idxi += ll
    return warped

def is_iterable(var):
    """
    Checks if the input is an iterable or not

    Parameters
    ----------
    var : integer, float, string, list

    Returns
    -------
    boolean
        Returns 'True' if var is iterable else False

    """

    try:
        for ii in var:
            return True
    except:
        return False
    return True


def force_iterable(var):
    """
    Forces var to be iterable, if not already

    Parameters
    ----------
    var : integer, float, string , list

    Returns
    -------
    iterable
        var as iterable

    """
    if not is_iterable(var):
        return [var]
    else:
        return var


def does_not_contain_strings(iterable):
    """
    Checks if iterable has any string element, returns False if it contains atleast one string

    Parameters
    ----------
    iterable : integer, float, string or any combination thereof

    Returns
    -------
    boolean
        True if iterable does not contain any string, else False

    """

    return all([not isinstance(ii, str) for ii in iterable])

def unique_list_of_iterables_by_tuple_hashing(ilist, return_idxs=False,
                                              ignore_order=False):
    """
    Returns the unique entries(if there are duplicates) from a list of iterables.

    Default is to take order into account, i.e. [[0,1],[1,0]]
    are considered different iterables

    If :obj:`ilist` contains non-iterables,
    they will be turned into iterables, s.t.
    1==[1]==np.array(1) and 'A'==['A'].
    They will also be returned as iterables

    Parameters
    ----------
    ilist : list of iterables
        list of iterables with redundant entries (redundant in the list, not in entries)
    return_idxs : boolean
        'True' if required to return indices instead of unique list. (Default is False).
    ignore_order : bool, default is False
        ignore order, s.t. [0,1] and [1,0]
        are considered equal. Only the first
        instance ([0,1]) is kept

    Returns
    -------
    result : list
        list of unique iterables or indices of 'ilist' where the unique entries are

    """

    # Check for stupid input
    if len(ilist) == 1:
        # TODO: avoid using print functions and use loggers
        print("Input is an iterable of len = 1. Doing nothing")
        if not return_idxs:
            return ilist
        else:
            return [0]

    if ignore_order:
        lambda_sort = lambda inlist : sorted(inlist)
    else:
        lambda_sort = lambda inlist : inlist
    # Now for the actual work
    idxs_out = []
    ilist_out = []
    seen = []
    for ii, sublist in enumerate(ilist):
        if isinstance(sublist, _np.ndarray):
            sublist = sublist.flatten()
        this_objects_id = hash(tuple(lambda_sort(force_iterable(sublist))))

        if this_objects_id not in seen:
            ilist_out.append(force_iterable(sublist))
            idxs_out.append(ii)
            seen.append(this_objects_id)
    if not return_idxs:
        return ilist_out
    else:
        return idxs_out

def window_average_fast(input_array_y, half_window_size=2):
    """
    Returns the moving average using np.convolve
    Parameters
    ----------
    input_array_y : array
                    numpy array for which moving average should be calculated
    half_window_size : int
                        the actual window size will be 2 * half_window_size + 1.
                        Example- when half window size = 2, moving average calculation will use window=5
    Returns
    -------
    array

    """
    input_array_y = (input_array_y).astype(float)
    window = _np.ones(2*half_window_size+1)
    return _np.convolve(input_array_y, window, mode="valid")/len(window)

#TODO consider list and str utils for this?
# from https://www.rosettacode.org/wiki/Range_expansion#Python
def rangeexpand(txt):
    """
    For a given integer range or multiple integer ranges, returns a list of individual integers.
    Example- "1-2,3-4" will return [1,2,3,4]

    Parameters
    ----------
    txt : string
        string of integers or integer range separated by ","

    Returns
    -------
    list
        list of integers

    """
    lst = []
    for r in txt.split(','):
        if '-' in r[1:]:
            r0, r1 = r[1:].split('-', 1)
            lst += range(int(r[0] + r0), int(r1) + 1)
        else:
            lst.append(int(r))
    return lst

# TODO CONSISIDER MERGING in_what_N_fragments w. in_what_fragment
def in_what_N_fragments(idxs, fragment_list):
    """
    For each element of idxs, return the index of "fragments" in which it appears

    Parameters
    ----------
    idxs : integer, float, or iterable thereof
    fragment_list : iterable of iterables
        iterable of iterables containing integers or floats

    Returns
    -------
    list
        list of length len(idxs) containing an iterable with the indices of 'fragments' in which that index appears

    """

    # Sanity checking
    idxs = force_iterable(idxs)
    assert does_not_contain_strings(idxs), ("Input 'idxs' cannot contain strings")
    # TODO: consider refactoring the loop into the function call (avoid iterating over function calls)
    assert all([does_not_contain_strings(ifrag) for ifrag in fragment_list]), ("Input 'fragment_list'  cannot contain strings")

    result = (_np.vstack([_np.in1d(idxs, iseg) for ii, iseg in enumerate(fragment_list)])).T
    return [_np.flatnonzero(row) for row in result]

def in_what_fragment(residx,
                     list_of_nonoverlapping_lists_of_residxs,
                     fragment_names=None):
    """
    For the residue id, returns the name(if provided) or the index of the "fragment"
    in which it appears

    Parameters
    ----------
    residx : int
        residue index
    list_of_nonoverlapping_lists_of_residxs : list
        list of integer list of non overlapping ids
    fragment_names : (optional) list of strings
        fragment names for each list in list_of_nonoverlapping_lists_of_residxs

    Returns
    -------
    integer or string
        returns the name(if names is provided) otherwise returns index of the "fragment"
        in which the residue index appears

    """

    # TODO deal with np.int64 etc
    assert type(residx) in (int, _np.int64), "Incorrect input: residx should be int, and not %s" % type(residx)
    assert does_not_contain_strings(list_of_nonoverlapping_lists_of_residxs), list_of_nonoverlapping_lists_of_residxs

    if fragment_names is not None:
        assert len(fragment_names) == len(list_of_nonoverlapping_lists_of_residxs)
    for ii, ilist in enumerate(list_of_nonoverlapping_lists_of_residxs):
        if residx in ilist:
            if fragment_names is None:
                return ii
            else:
                return fragment_names[ii]

def exclude_same_fragments_from_residx_pairlist(pairlist,
                                                fragments,
                                                return_excluded_idxs=False):
    """
    If the members of the pair belong to the same fragment, exclude them from pairlist.

    Parameters
    ----------
    pairlist : list of iterables
        each iterable within the list should be a pair.
    fragments : list of iterables
        each inner list should have residue indexes that form a fragment
    return_excluded_idxs : boolean
        True if index of excluded pair is needed as an output. (Default is False).

    Returns
    -------
    list
        pairs that don't belong to the same fragment,
        or index of the excluded pairs if return_excluded_idxs is True

    """

    idxs2exclude = [idx for idx, pair in enumerate(pairlist) if
                    _np.diff([in_what_fragment(ii, fragments) for ii in pair]) == 0]

    if not return_excluded_idxs:
        return [pair for ii, pair in enumerate(pairlist) if ii not in idxs2exclude]
    else:
        return idxs2exclude

def assert_min_len(input_iterable, min_len=2):
    """
    Checks if an iterable satisfies the criteria of minimum length. (Default minimum length is 2).
    Parameters
    ----------
    input_iterable : numpy array, list of list
                    example np.zeros((2,1,1) or [[1,2],[3,4]] when min_len = 2
    min_len : minimum length which the iterable should satisfy (Default is 2)

    Returns
    -------
    Prints error if each item within the iterable has lesser number of elements than min_len

    """
    for ii, element in enumerate(input_iterable):
        if _np.ndim(element)==0 or len(element) < min_len:
            aerror = 'The %s-th element has too few elements (min %s): %s' % (ii, min_len, element)
            raise AssertionError(aerror)

def join_lists(lists, idxs_of_lists_to_join):
    r"""
    Provided a list of lists, join them following idxs_of_lists_to_join

    Parameters
    ----------
    lists: iterable of iterables
        The lists to be joined

    idxs_of_lists_to_join: iterable of iterables containing integers
        The lists to join. These  3 things will be done before using this array
            - remove duplicate entries in each iterable
            - sort the entries in each iterable by ascending order
            - assert there is no overlap between iterables

    Returns
    -------
    joined_lists: iterable of iterables
        :obj:`lists` joined following the criterion of :obj:`idxs_of_lists_to_join`
        Once the new iterables have been created by joining the initial interables,
        they will be re-ordered by ascending first element

    """
    # Check that the input is an iterable of iterables
    assert_min_len(idxs_of_lists_to_join)

    # Removing the redundant entries in each list and sorting them
    idxs_of_lists_to_join = [_np.unique(jo) for jo in idxs_of_lists_to_join]

    # Assert the join_fragments do not overlap
    assert_no_intersection(idxs_of_lists_to_join)
    joined_lists = []
    lists_idxs_used_for_joining = []
    for ii, jo in enumerate(idxs_of_lists_to_join):
        # print(ii,jo)
        this_new_frag = []
        lists_idxs_used_for_joining.extend(jo)
        for frag_idx in jo:
            # print(frag_idx)
            this_new_frag.extend(lists[frag_idx])
        # print(this_new_frag)
        joined_lists.append(_np.array(this_new_frag))

    # TODO: THIS is only good programming bc the lists are very very small, otherwise np.delete is the way to go
    surviving_initial_fragments = [ifrag for ii, ifrag in enumerate(lists)
                                   if ii not in lists_idxs_used_for_joining]
    lists = joined_lists + surviving_initial_fragments

    # Order wrt to the first index in each fragment
    order = _np.argsort([ifrag[0] for ifrag in lists])
    lists = [lists[oo] for oo in order]

    return lists

def assert_no_intersection(list_of_lists_of_integers, word='iterables'):
    """
    Checks if two or more lists contain the same integer
    Parameters
    ----------
    list_of_lists_of_integers : list of lists

    Returns
    -------
    Prints assertion message if inner lists have the same integer, else no output

    """
    # Looping here because the number of fragments will never become too large
    # Also of interest
    # len(f1),len(f2)
    # (196, 709)
    # %timeit np.intersect1d(f1,f2);
    # 24.4 µs ± 1.47 µs per loop (mean ± std. dev. of 7 runs, 10000 loops each)

    # %timeit set(f1).intersection(f2);
    # 84.9 µs ± 1.95 µs per loop (mean ± std. dev. of 7 runs, 10000 loops each)

    for ii, jj in _np.vstack(_np.triu_indices(len(list_of_lists_of_integers), k=1)).T:
        l1, l2 = list_of_lists_of_integers[ii], list_of_lists_of_integers[jj]
        assert len(l1) != 0 or len(l2) != 0, (l1,l2, "Both lists are empty! See https://www.coopertoons.com/education/emptyclass_intersection/emptyclass_union_intersection.html")
        assert len(_np.intersect1d(l1, l2)) == 0, 'input %s %u and %u have these elements ' \
                                                          'in common: %s:\n%s\nvs\n%s'%(word, ii, jj, set(l1).intersection(l2), l1, l2)

def put_this_idx_first_in_pair(idx, pair):
    """
    Returns the original pair if the value already appears first, else returns reversed pair
    Parameters
    ----------
    idx : value which needs to be brought in the first place (not the index but value itself)
    pair : list
            pair of values as a list
    Returns
    -------
    pair

    """
    if pair[0] != idx and pair[1] == idx:
        pair = pair[::-1]
    elif pair[0] == idx and pair[1] != idx:
        pass
    else:
        print(pair)
        raise Exception
    return pair

def hash_list(ilist):
    r"""Try to hash all the objects of a list (regardless of type) into one hash

    Parameters
    ----------
    iobj : anthing

    Returns : hashed object
    -------

    """
    list_of_hashes = []
    for iobj in ilist:
        try:
            res = hash(iobj)
        except:
            try:
                res = hash(tuple(iobj))
            except:
                try:
                    res = hash(tuple([hash(tuple(subitem)) for subitem in iobj]))
                except:
                    try:
                        res = hash(tuple([hash(tuple(subitem.flatten())) for subitem in iobj]))
                    except Exception as e:
                        print("Cannot hash type", type(iobj))
                        raise e
        list_of_hashes.append(res)
    return hash(tuple(list_of_hashes))

def idx_at_fraction(val_desc_order, frac):
    r"""
    Index of :obj:`val_desc_order` where np.cumsum(val)/np.sum(val)>= frac for the first time

    Parameters
    ----------
    val_desc_order : array like of floats
        The values that the determine the sum of which a fraction will be taken
        The have to be in descending order
    frac : float
        The target fraction of sum(val) that is needed

    Returns
    -------
    n : int
        Index of val where the fraction is attained for the first time.
        For the number of entries of :obj:`val`, just use n+1
    """
    assert all(_np.diff(val_desc_order)<=0), "Values must be in descending order!"
    assert 0<=frac<=1, "Fraction has to be in [0,1] ,not %s"%frac
    normalized_cumsum = (_np.cumsum(val_desc_order) / _np.sum(val_desc_order)).round(5) >= frac
    #The rounding factor can be significant if many contacts are involved,
    # the numercal comparison cannot be 1.0000000000000 because it might be never reached.
    # Previsouly it was round(2) but yields errors for very long lists of contacts
    return _np.flatnonzero(normalized_cumsum>=frac)[0]


def _get_n_ctcs_from_freqs(ctc_control, ctc_freqs):
    r"""
    Helper method to understand what :obj:`ctc_control` is meant to do with :obj:`ctc_freqs`

    Provided with a set of frequencies :obj:`ctc_freqs`, the user decides how
    many them are kept by using the parameter `ctc_control`

    Parameters
    ----------
    ctc_control : int or float
        * If int:
          Keep these many contacts, i.e.
          return ctc_control directly
        * If float:
          Interpret ctcs_freq control as
          a fraction [0,1] and return
          cumsum(ctc_freqs)/sum(ctc_freqs)>= ctc_control
          That gives the number of contacts needed
          to "keep" the fraction of frequencies.
          Check :obj:`idx_at_fraction` for more info
    ctc_freqs : iterable
        Floats in descending order
    Returns
    -------
    n_ctcs : int
    or_fraction_needed : bool
        Whether an orienting-fraction value
        will be needed downstream. Is
        True if ctc_control was an integer
        and False if it was a float (if
        it was a float, we were interested in a fraction
        anyways)
    """
    or_fraction_needed = True
    total_n_ctcs = _np.array(ctc_freqs).sum()
    if isinstance(ctc_control, int):
        n_ctcs = int(ctc_control)
    else:
        if total_n_ctcs > 0:
            n_ctcs = idx_at_fraction(ctc_freqs, ctc_control) + 1
            or_fraction_needed = False
        else:
            n_ctcs = 0
    return n_ctcs, or_fraction_needed

def remove_from_lists(list_of_lists, remove_these):
    r"""
    Wraps safely around :obj:`numpy.setdiff1d` not returning empty lists

    Parameters
    ----------
    list_of_lists : iterable of iterables
    remove_these : iterable

    Returns
    -------
    clean_list : list
    """
    _fragments = [_np.setdiff1d(fr, remove_these, assume_unique=True) for fr in list_of_lists]
    return [fr.tolist() for fr in _fragments if len(fr)>0]

def find_parent_list(sublists, parent_lists):
    r"""
    For each sublist, return the index of the parent list

    Parameters
    ----------
    sublists : list of iterables
    parent_lists : list of iterables

    Returns
    -------
    parents_by_child : list
        A list of len(sublists) with indices
        indicating which element of :obj:`parent_lists`
        each sublist is a subset of. If a sublist
        doesn't have a parent, its parent is None
    child_by_parent : dict
        A dictionary keyed by parent idx
        and valued with idxs of their children
    """
    assert_no_intersection(parent_lists)
    parents_by_child=[]
    child_by_parent={}
    for sf in sublists:
        parent = None
        iset = set(list(sf))
        for pp, par in enumerate(parent_lists):
            if set(list(par)).issuperset(iset):
                parent = pp
                break
        parents_by_child.append(parent),
    for ii, __ in enumerate(parent_lists):
        kids = _np.flatnonzero(_np.array(parents_by_child)==ii)
        if len(kids):
            child_by_parent[ii]=kids
    return parents_by_child, child_by_parent

def unique_product_w_intersection(a1,a2):
    r"""
    Fast way to create the product of two intersecting sets without repeated/unwanted pairs

    Consider that
    >>> list(itertools.product([0,1,2,3],[2,3,4,5]))
    [(0, 2),
     (0, 3),
     (0, 4),
     (0, 5),
     (1, 2),
     (1, 3),
     (1, 4),
     (1, 5),
     (2, 2),
     (2, 3),
     (2, 4),
     (2, 5),
     (3, 2),
     (3, 3),
     (3, 4),
     (3, 5)]

    Has the repeated/unwanted pairs (2,2),(3,3),(3,2) which
    need to be taken out a posteriori by comparing pairs.

    The :obj:`unique_list_of_iterables_by_tuple_hashing`
    method accepts also arrays (since pairlists
    may not necessarily have been generated
    as tuples, but also as np.arrays), s.t.
    the arrays need to be casted into tuples before hashing
    and one comparison per pair (grows quadratically)

    >>> a1 = np.arange(200)
    >>> a2 = np.arange(195,300)
    >>> pairs = np.array(list(itertools.product(a1,a2)))
    >>> %timeit mdciao.utils.lists.unique_list_of_iterables_by_tuple_hashing(slow)
    2.83 s ± 170 ms per loop (mean ± std. dev. of 7 runs, 1 loop each)

    Whereas
    >>> %timeit mdciao.utils.lists.unique_product_w_intersection(a1,a2)
    47 ms ± 394 µs per loop (mean ± std. dev. of 7 runs, 10 loops each)


    For reference
    >>> %timeit list(itertools.product(a1,a2))
    783 µs ± 5.37 µs per loop (mean ± std. dev. of 7 runs, 1000 loops each)

    I.e. clearly, for non-intersecting sets a1 and a2 without unwanted/repeated
    pairs, it's always better to use itertools.product directly

    Parameters
    ----------
    a1 : iterable
        The integers of the set1
    a2 : iterable
        The integers of the set2

    Returns
    -------
    pairlist : np.ndarray
        The pairlist product of a1 and a2
        without self-pairs (ii,ii) and the
        only (ii,jj) (not (jj,ii))

    """
    from itertools import product, combinations
    intersect = list(set(a1).intersection(a2))
    a1_no_int = list(set(a1).difference(intersect))
    a2_no_int = list(set(a2).difference(intersect))
    pairlist = list(product(a1_no_int, a2_no_int))+list(product(a1_no_int,intersect))\
                                                  +list(product(intersect,a2_no_int))\
               +list(combinations(intersect,2))
    pairlist = _np.vstack(sorted(pairlist, key=lambda item: (item[0], item[1]), reverse=False))
    return pairlist



r"""
Welcome to mdciao! Below you will find the submodules exposed by the API

.. autosummary::
   :toctree: generated/

    contacts
    fragments
    nomenclature
    sites
    cli
    plots
    utils
    flare
    pdb
    examples
"""

from . import contacts
from . import utils
from . import fragments
from . import plots
from . import sites
from . import cli
from . import flare
from . import pdb
from . import examples


##############################################################################
#    This file is part of mdciao.
#    
#    Copyright 2020 Charité Universitätsmedizin Berlin and the Authors
#
#    Authors: Guillermo Pérez-Hernandez
#    Contributors:
#
#    mdciao is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    mdciao is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with mdciao.  If not, see <https://www.gnu.org/licenses/>.
##############################################################################

from os import path as _path, getcwd as _getcwd, chdir as _chdir, link as _link, mkdir as _mkdir
from shutil import copy as _shcopy
from glob import glob as _glob
from mdciao import __path__ as mdc_path
from subprocess import run as _run
from zipfile import ZipFile as _ZF
from mdciao.filenames import FileNames as _FN

filenames = _FN()
mdc_path = _path.split(mdc_path[0])[0]

long2short = {"--residues" : "-r",
              "--n_smooth_hw" : "-ns",
              "--table_ext" : "-tx",
              "--GPCR_uniprot" : "--GPCR",
              "--CGN_PDB"   : "--CGN"
              }

long2long = {key:key for key in long2short.keys()}

from requests import get as _rget
from tqdm.auto import tqdm as _tqdma
import contextlib as _contextlib
from mdciao.cli import residue_neighborhoods as _residue_neighborhoods, interface as _interface
from mdciao.nomenclature import LabelerGPCR as _LabelerGPCR, LabelerCGN as _LabelerCGN, LabelerKLIFS as _LabelerKLIFS
from tempfile import TemporaryDirectory as _TDir, TemporaryFile as _TF
import io as _io
@_contextlib.contextmanager
def remember_cwd():
    curdir = _getcwd()
    try:
        yield
    finally:
        _chdir(curdir)

class ExamplesCLTs(object):
    def __init__(self, test=False, short=False):
        r"""
        API interface to show or call the mdc_* scripts from the command line directly

        For instance, self.run("mdc_interface") will run an mdc_interface.py example

        Parameters
        ----------
        test : bool, default is False
            If True, internally all references to input filenames
            are converted from full paths to basenames, because
            it is assumed that the test method is calling the
            self.run() method form a temporary directory
            where links to these filenames have been
            created already. Additionally, the flag
            "-ni" for non-interactive runs is added to the methods
            that would otherwise prompt the user for some confirmations
        short : bool, default is False
            Whether to use the short or the long versions
            of the flags
        """
        self.xtc = filenames.traj_xtc
        self.pdb = filenames.top_pdb
        self.GPCRlabs_file = filenames.adrb2_human_xlsx
        self.KLIFSlabs_file = filenames.KLIFS_P31751_xlsx
        self.CGN_file = filenames.CGN_3SN6
        self.sitefile = filenames.tip_json
        self.pdb_3SN6 = filenames.pdb_3SN6
        self.KLIFS_pdb = filenames.pdb_3E8D

        self.test = test
        cwd = _getcwd()
        for fn in ["xtc", "pdb", "GPCRlabs_file", "CGN_file", "sitefile", "pdb_3SN6"]:
            attr_val = getattr(self, fn)
            if self.test:
                setattr(self, fn, _path.basename(attr_val))
            else:
                setattr(self, fn, _path.relpath(attr_val,cwd))

        if short:
            self.opt_dict=long2short
        else:
            self.opt_dict = long2long
    @property
    def mdc_neighborhoods(self):
        return ["mdc_neighborhoods.py",
                "%s %s" % (self.pdb, self.xtc),
                self.opt_dict["--residues"] + " L394",
                self.opt_dict["--n_smooth_hw"] + " 1",
                self.opt_dict["--table_ext"] + " xlsx",
                self.opt_dict["--GPCR_uniprot"] + " %s" % self.GPCRlabs_file,
                self.opt_dict["--CGN_PDB"] + " %s" % self.CGN_file,
                ]
    @property
    def mdc_sites(self):
        return ["mdc_sites.py ",
                "%s %s" % (self.pdb, self.xtc),
                " --site_files %s" % self.sitefile,
                " --GPCR_uniprot %s" % self.GPCRlabs_file,
                " --CGN_PDB %s" % self.CGN_file
                ]

    @property
    def mdc_interface(self):
        return ["mdc_interface.py ",
                "%s %s" % (self.pdb, self.xtc),
                " --frag_idxs_group_1 0-2",
                " --frag_idxs_group_2 3",
                " --ctc_control 20",
                " --GPCR_uniprot %s" % self.GPCRlabs_file,
                " --CGN_PDB %s" % self.CGN_file,
                ]
    @property
    def mdc_GPCR_overview(self):
        return ["mdc_GPCR_overview.py",
                "%s" % self.GPCRlabs_file,
                "-t %s" % self.pdb]

    @property
    def mdc_KLIFS_overview(self):
        return ["mdc_KLIFS_overview.py",
                "%s" % self.KLIFSlabs_file,
                "-t %s" % self.KLIFS_pdb]

    @property
    def mdc_CGN_overview(self):
        # This is the only one that needs network access
        return ["mdc_CGN_overview.py",
                "%s" % '3SN6',
                "-t %s" % self.pdb,
                ]

    @property
    def mdc_compare(self):
        return ["mdc_neighborhoods.py",
                "%s %s" % (self.pdb, self.xtc),
                self.opt_dict["--residues"] + " L394",
                self.opt_dict["--n_smooth_hw"] + " 1",
                self.opt_dict["--table_ext"] + " xlsx",
                "--ctc_cutoff_Ang 3",
                "\n\n",
                "mdc_neighborhoods.py",
                "%s %s" % (self.pdb, self.xtc),
                self.opt_dict["--residues"] + " L394",
                self.opt_dict["--n_smooth_hw"] + " 1",
                self.opt_dict["--table_ext"] + " xlsx",
                "--ctc_cutoff_Ang 4",
                "\n\n"
                "mdc_compare.py",
                "neighborhood.LEU394@frag0@3.0_Ang.xlsx",
                "neighborhood.LEU394@frag0@4.0_Ang.xlsx"
                ]
        pass

    @property
    def mdc_fragments(self):
        return ["mdc_fragments.py ",
                "%s" % (self.pdb)
                ]
        pass

    @property
    def mdc_pdb(self):
        return ["mdc_pdb.py 3SN6"]

    @property
    def mdc_residues(self):
        return ["mdc_residues.py ",
                "P0G,380-394,3.5* "
                "%s"% (self.pdb),
                " --GPCR_uniprot %s" % self.GPCRlabs_file,
                "-ni"]

    @property
    def mdc_notebooks(self):
        return ["mdc_notebooks.py "]

    @property
    def clts(self):
        return [attr for attr in dir(self) if attr.startswith("mdc")]

    def _assert_clt_exists(self, clt):
        assert clt in self.clts, "Input method %s is not in existing methods %s" % (clt, self.clts)

    def _join_args(self,clt):
        oneline = self.__getattribute__(clt)
        # Only add the -ni argument where it's needed
        if self.test and all([istr not in clt for istr in ["_overview", "mdc_fragments", "mdc_pdb", "mdc_compare", "mdc_notebooks"]]):
            oneline += ["-ni"]
        return " ".join(oneline)

    def show(self, clt):
        self._assert_clt_exists(clt)
        print("%s example call:" % clt)
        print("%s--------------"%("".join(["-" for _ in clt]))) # really?
        oneline = self._join_args(clt)
        print(oneline.replace(" -", " \n-"))
        print("\n\nYou can re-run 'mdc_examples %s.py' with the '-x' option to execute the command directly\n"
              "or you can paste the line below into your terminal, add/edit options and then execute:\n"%clt)
        print(oneline)

    def run(self, clt,show=True):
        if show:
            self.show(clt)
        oneline = self._join_args(clt)
        CP = []
        for line in oneline.split("\n\n"):
            CP.append(_run(line.split()))
        if self.test:
            return CP


def ContactGroupL394(**kwargs):
    r"""
    Create an example :obj:`mdciao.contacts.ContactGroup` quickly.

    Wraps around :obj:`mdciao.cli.residue_neighborhoods` and asks
    for the LEU394 neighborhood.

    The input data is one very short (80 frames) version
    of the MD trajectory shipped with mdciao, kindly provided by
    Dr. H. Batebi. See the online examples for more info.

    Parameters
    ----------
    kwargs : optional keyword arguments
        For :obj:`mdciao.cli.residue_neighborhoods`

    Returns
    -------
    CG : a :obj:`~mdciao.contacts.ContactGroup`

    """
    # TODO make a method out of this link+cd_tmpdir+return
    with _TDir(suffix="_mdciao_example_CG") as t:
        for fn in [filenames.pdb_3SN6, filenames.traj_xtc,
                   filenames.top_pdb,
                   filenames.adrb2_human_xlsx, filenames.CGN_3SN6]:
            _link(fn, _path.join(t, _path.basename(fn)))

        with remember_cwd():
            _chdir(t)
            b = _io.StringIO()
            try:
                with _contextlib.redirect_stdout(b):
                    example_kwargs = {"topology": _path.basename(filenames.top_pdb),
                                      "n_smooth_hw": 1,
                                      "figures": False,
                                      "GPCR_uniprot": _path.basename(filenames.adrb2_human_xlsx),
                                      "CGN_PDB": _path.basename(filenames.CGN_3SN6),
                                      "no_disk":True,
                                      "accept_guess": True}
                    for key, val in kwargs.items():
                        example_kwargs[key]=val
                    return _residue_neighborhoods("L394",
                                                  _path.basename(filenames.traj_xtc),
                                                  **example_kwargs,
                                                  )["neighborhoods"][353]

            except Exception as e:
                print(b.getvalue())
                b.close()
                raise e

def notebooks(folder ="mdciao_notebooks"):
    r"""
    Copy the example Jupyter notebooks distributed with mdciao to this folder.

    The method never overwrites an existing folder, but
    keeps either asking or producing new folder names.

    Parameters
    ----------
    folder : str

    Returns
    -------
    folder : str
        The folder to which the notebooks were copied
        Can be identical to the input or one generated
        by :obj:`_recursive_prompt`.

    """

    pwd = _getcwd()

    nbs = sorted(_glob(_path.join(filenames.notebooks_path,"*.ipynb")))

    dest = _recursive_prompt(_path.join(pwd, folder), folder, verbose=True)
    print("Copying file(s):"
          "\n -%s"%"\n -".join(nbs))
    _mkdir(dest)
    print("Here:")
    for ff in nbs:
        _shcopy(ff,dest)
        print(" %s"%_path.join(dest,_path.basename(ff)))

    return dest

def _recursive_prompt(input_path, pattern, count=1, verbose=False, is_file=False):
    r"""
    Ensure input_path doesn't exist and keep generating/prompting for alternative filenames/dirnames


    Poorman's attempt at actually a good recursive function, but does its job.
    A maximum recursion depth of 50 is hard-coded

    Parameters
    ----------
    input_path : str
        Example, "path/to/mdicao_example.zip"
    pattern : str
        The part of input_path used to generate new filenames,
        example "mdciao_example"
    count : int, default is 0
        Where in the recursion we are
    verbose : bool, default is False
    is_file : book, default is False
        Name-generating is different for folders than
        from files:
        * mdciao_notebook -> mdciao_notebook_00
        * mdciao_example.zip -> mdciao_example_00.zip

    Returns
    -------
    nox_path : str
        A newly created, previosuly non-existent path

    """

    max_recurr = 50
    cwd = _getcwd()
    ext = _path.splitext(input_path)[1]
    while _path.exists(input_path):
        if verbose:
            print("%s exists" % input_path)
        input_path = _path.join(cwd, pattern) + "_%02u" % count
        if is_file:
            input_path += ext
        count += 1
        if count > max_recurr:
            raise RecursionError("Over %u files/folders exist with the '%s' pattern, aborting" % (max_recurr, pattern))

    if count>1:
        print(input_path, "will be created")
        print("Hit Enter to accept or provide another path from %s%s:\r"%(cwd,_path.sep))
        answer = input()
    else:
        answer = ""
    if len(answer)==0:
        return input_path
    else:
        input_path = _path.join(cwd, answer)
        print("OK, your suggestion is",input_path)
    if _path.exists(input_path):
        print("%s already exists. Next suggestion:" % input_path)
        return _recursive_prompt(input_path, pattern, count=count, verbose=verbose, is_file=is_file)
    else:
        return _path.join(cwd,answer)


def fetch_example_data(alias_or_url="b2ar@Gs",
                       unzip=True):
    r""" Download the example data as zipfile and unzip it to the working directory.

    This data is used in the notebooks:
     * Manuscript.ipynb (b2ar@Gs)
     * Tutorial.ipynb (b2ar@Gs)
     * Missing_Contacts.ipynb (b2ar@Gs)
     * EGFR Kinase Inhibitors.ipynb (EGFR)
     * Comparing_CGs_Bars.ipyn (cov19)
     * Comparing_CGs_Flares.ipynb (cov19)
    which can all be run locally issuing,
    from the CLI:

    >>>  mdc_notebooks.py

    or from the API:

    >>> mdciao.examples.notebooks()

    Note
    ----
    New filenames for the downloaded file, and the resulting folder
    will be generated to avoid overwriting. No files will
    be overwritten when extracting.

    Parameters
    ----------
    alias_or_url : str
        The url or the alias to download example data.
        Currently, these are the available aliases and their urls
         * b2ar@Gs : https://proteinformatics.uni-leipzig.de/mdciao/mdciao_example.zip
          Beta 2 adrenergic receptor in complex with Gs-protein. Provided
          kindly by H. Batebi (1 traj, ca. 10 MB, 280 frames, dt = 10 ps)

         * EGFR : http://proteinformatics.uni-leipzig.de/mdciao/example_kinases.zip
          Epidermal Growth Factor Receptor (EGFR) in complex with
          four inhibitors. The inhibitors and their PDB IDs are
          P31@3POZ, W321@3W32, EUX1@6LUB and  7VH1@7VRE.
          The data has been generated using `TeachOpenCADD <https://projects.volkamerlab.org/teachopencadd/index.html>`_ .
          (4 trajs, ca 10 MB each, ca 500 frames each, dt = 1ns)

         * cov19 : https://proteinformatics.uni-leipzig.de/mdciao/example_cov19.zip
          SARS-CoV-2 spike protein receptor binding domain (RBD) bound
          to human angiotensin converting enzyme-related carboypeptidase (ACE2).
          The files already contain processed trajectory data, in the form
          of mdciao.contacts.ContactGroup-objects stored as npy files.
          Original data generated at the Chodera Lab by Ivy Zhang,
          made available via `molSSI <https://covid.molssi.org//simulations/#foldinghome-simulations-of-the-sars-cov-2-spike-rbd-bound-to-human-ace2>`_.
          (1 npy file with interfaces for 4 setups and one sample trajectory file, ca 35 MB)

    unzip : bool, default is True
        Try unzipping the file after downloading

    Returns
    -------
    downed_fold_full : str
        The full path to the downloaded data

    """
    # TODO change to proteinformatics.org when ssl problems get fixed
    alias2url = {"b2ar@Gs": "https://proteinformatics.uni-leipzig.de//mdciao/mdciao_example.zip",
                 "EGFR": "https://proteinformatics.uni-leipzig.de/mdciao/example_kinases.zip",
                 "cov19" : "https://proteinformatics.uni-leipzig.de/mdciao/example_cov19.zip",
                 "test": "https://proteinformatics.uni-leipzig.de/mdciao/mdciao_test_small.zip"}

    if alias_or_url in alias2url.keys():
        url = alias2url[alias_or_url]
    elif alias_or_url in alias2url.values():
        url = alias_or_url
    else:
        raise ValueError("Cannot find %s in the known aliases or in the known urls:\n%s" % (alias_or_url, alias2url))
    downed_file_full = _down_url_safely(url)
    if unzip:
        return _unzip2dir(downed_file_full)
    else:
        return  downed_file_full

def _unzip2dir(full_path_zipfile):
    r"""
    Unzip to a folder with the same name as the file, regardless of the structure of the zipfile

    The folder's full path is kept, including zipfile's name minus the .zip extension

    Background: "mdciao_example.zip" was zipped in origin with this structure:
     * mdciao_example/prot.pdb
     * mdciao_example/traj.xtc

    However, it might have been renamed to "mdciao_example_05.zip" when auto-downloading.
    If you unzip it directly, it will be extracted to mdciao_example, which ideally
    should be avoided.

    If for whatever reason, mdciao_example_05 *does* exist as a folder, existing files
    will not be overwritten

    Parameters
    ----------
    full_path_zipfile : str
        The zipfile. E.g. "Downloads/mdciao_example_00.zip" will be
        extracted to "Downloads/mdciao_example_00/"

    Returns
    -------
    full_dir : str
        The directory into which the zipfile was extracted

    """
    full_dir = _path.splitext(full_path_zipfile)[0]
    # https://stackoverflow.com/a/56362289
    #local_dir = _path.basename(full_dir)
    print("Unzipping to '%s'" % full_dir)
    with _ZF(full_path_zipfile) as zipdata:
        for zipinfo in zipdata.infolist():
            new = _path.basename(zipinfo.filename)
            zipinfo.filename = new
            if _path.exists(_path.join(full_dir,new)):
                print("No unzipping of '%s': file already exists." % new)
            else:
                zipdata.extract(zipinfo,
                                path=full_dir
                                )
            #assert _path.exists(_path.join(full_dir,new))

    return full_dir

def _down_url_safely(url, chunk_size = 128, verbose=False):
    r"""
    Downloads a file from a URL to a tmpfile and copies it to the current directory

    If the file/folder exists already, a new filename is generated using _recursive_prompt

    Parameters
    ----------
    url
    chunk_size

    Returns
    -------

    """
    filename_orig = _path.basename(url)
    filename_nonx = _recursive_prompt(filename_orig,
                                      _path.splitext(filename_orig)[0],
                                      is_file=True, verbose=True)
    r = _rget(url, stream=True)
    total_size_in_bytes = int(r.headers.get('content-length', 0))
    pb = _tqdma(total=total_size_in_bytes,
                desc="Downloading %s to %s" % (filename_orig, _path.basename(filename_nonx)))
    with _TDir(suffix="_mdciao_download") as t:
        _filename = _path.join(t,_path.basename(filename_nonx))
        with open(_filename, 'wb') as fd:
            r.iter_content()
            for chunk in r.iter_content(chunk_size=chunk_size):
                fd.write(chunk)
                pb.update(128)
        pb.close()
        if verbose:
            print("Dowloaded file %s"%filename_nonx)
        _shcopy(_filename,filename_nonx)

    return filename_nonx

def GPCRLabeler_ardb2_human(**kwargs):
    r"""Build an :obj:`~mdciao.nomenclature.LabelerGPCR` with the adrb2_human.xlsx file shipped with mdciao"""
    return _LabelerGPCR(filenames.adrb2_human_xlsx,**kwargs)

def CGNLabeler_3SN6(**kwargs):
    r"""Build an :obj:`~mdciao.nomenclature.LabelerCGN` with the CGN_3SN6.txt and 3SN6.pdb files shipped with mdciao"""
    with _TDir(suffix="_mdciao_example_CGNLabeler") as t:
        for fn in [filenames.pdb_3SN6, filenames.CGN_3SN6]:
            _link(fn, _path.join(t, _path.basename(fn)))
        with remember_cwd():
            _chdir(t)
            CGN = _LabelerCGN("3SN6",**kwargs)
    return CGN

def Interface_B2AR_Gas(**kwargs):
    r"""
    Return an :obj:`~mdciao.contacts.ContactGroup` object representing a B2AR-Galpha s interface

    Wraps around :obj:`mdciao.cli.interface`

    The input data is one very short (80 frames) version
    of the MD trajectory shipped with mdciao, kindly provided by
    Dr. H. Batebi. See the online examples for more info.

    Parameters
    ----------
    kwargs : keyword args for :obj:`mdciao.cli.interface`

    Returns
    -------
    intf :obj:`~mdciao.contacts.ContactGroup`
    """
    b = _io.StringIO()
    try:
        with _contextlib.redirect_stdout(b):
            example_kwargs = {"topology": filenames.top_pdb,
                              "figures": False,
                              "GPCR_uniprot": GPCRLabeler_ardb2_human(),
                              "CGN_PDB": CGNLabeler_3SN6(),
                              "no_disk": True,
                              "frag_idxs_group_1":[0],
                              "frag_idxs_group_2":[3],
                              "ctc_control":1.0,
                              "accept_guess": True}
            for key, val in kwargs.items():
                example_kwargs[key] = val
            return _interface(filenames.traj_xtc,
                              **example_kwargs,
                              )

    except Exception as e:
        print(b.getvalue())
        b.close()
        raise e

def KLIFSLabeler_P31751():
    r"""Build an :obj:`~mdciao.nomenclature.LabelerKLIFS` with the KLIFS_P31751.xlsx and 3E8D.pdb.gz.pdb files shipped with mdciao"""
    with _TDir(suffix="_mdciao_example_KLIFS") as t:
        for fn in [filenames.KLIFS_P31751_xlsx, filenames.pdb_3E8D]:
            _link(fn, _path.join(t, _path.basename(fn)))

        with remember_cwd():
            _chdir(t)
            return _LabelerKLIFS("P31751", try_web_lookup=False)